"""
REST API handlers for RBAC (Role-Based Access Control) system.
Provides CRUD endpoints for role management and role assignment operations.
"""

from __future__ import annotations

from collections.abc import Iterable
from http import HTTPStatus

import aiohttp_cors
from aiohttp import web

from ai.backend.common.api_handlers import APIResponse, BodyParam, PathParam, api_handler
from ai.backend.common.contexts.user import current_user
from ai.backend.common.dto.manager.rbac import (
    AssignRoleRequest,
    AssignRoleResponse,
    CreateRoleRequest,
    CreateRoleResponse,
    DeleteRoleResponse,
    GetRolePathParam,
    GetRoleResponse,
    PaginationInfo,
    RevokeRoleRequest,
    RevokeRoleResponse,
    SearchRolesRequest,
    SearchRolesResponse,
    SearchUsersAssignedToRolePathParam,
    SearchUsersAssignedToRoleRequest,
    SearchUsersAssignedToRoleResponse,
    UpdateRolePathParam,
    UpdateRoleRequest,
    UpdateRoleResponse,
)
from ai.backend.common.dto.manager.rbac.path import SearchEntitiesPathParam, SearchScopesPathParam
from ai.backend.common.dto.manager.rbac.request import (
    DeleteRoleRequest,
    PurgeRoleRequest,
    SearchEntitiesRequest,
    SearchScopesRequest,
)
from ai.backend.common.dto.manager.rbac.response import (
    GetEntityTypesResponse,
    GetScopeTypesResponse,
    SearchEntitiesResponse,
    SearchScopesResponse,
)
from ai.backend.manager.api.auth import auth_required_for_method
from ai.backend.manager.api.types import CORSOptions, WebMiddleware
from ai.backend.manager.data.permission.role import UserRoleAssignmentInput, UserRoleRevocationInput
from ai.backend.manager.dto.context import ProcessorsCtx
from ai.backend.manager.errors.permission import NotEnoughPermission
from ai.backend.manager.models.rbac_models.role import RoleRow
from ai.backend.manager.repositories.base import Creator, Purger, Updater
from ai.backend.manager.repositories.permission_controller.creators import RoleCreatorSpec
from ai.backend.manager.services.permission_contoller.actions import (
    AssignRoleAction,
    CreateRoleAction,
    DeleteRoleAction,
    GetRoleDetailAction,
    RevokeRoleAction,
    SearchRolesAction,
    SearchUsersAssignedToRoleAction,
    UpdateRoleAction,
)
from ai.backend.manager.services.permission_contoller.actions.get_entity_types import (
    GetEntityTypesAction,
)
from ai.backend.manager.services.permission_contoller.actions.get_scope_types import (
    GetScopeTypesAction,
)
from ai.backend.manager.services.permission_contoller.actions.purge_role import PurgeRoleAction
from ai.backend.manager.services.permission_contoller.actions.search_entities import (
    SearchEntitiesAction,
)
from ai.backend.manager.services.permission_contoller.actions.search_scopes import (
    SearchScopesAction,
)

from .assigned_user_adapter import AssignedUserAdapter
from .entity_adapter import EntityAdapter
from .role_adapter import RoleAdapter
from .scope_adapter import ScopeAdapter

__all__ = ("create_app",)


class RBACAPIHandler:
    """REST API handler class for RBAC operations."""

    def __init__(self) -> None:
        self.role_adapter = RoleAdapter()
        self.assigned_user_adapter = AssignedUserAdapter()
        self._scope_adapter = ScopeAdapter()
        self._entity_adapter = EntityAdapter()

    # Role Management Endpoints

    @auth_required_for_method
    @api_handler
    async def create_role(
        self,
        body: BodyParam[CreateRoleRequest],
        processors_ctx: ProcessorsCtx,
    ) -> APIResponse:
        """Create a new role."""
        processors = processors_ctx.processors
        me = current_user()
        if me is None or not me.is_superadmin:
            raise NotEnoughPermission("Only superadmin can create roles.")

        # Convert request to creator
        creator = Creator(
            spec=RoleCreatorSpec(
                name=body.parsed.name,
                source=body.parsed.source,
                status=body.parsed.status,
                description=body.parsed.description,
            )
        )

        # Call service action
        action_result = await processors.permission_controller.create_role.wait_for_complete(
            CreateRoleAction(creator=creator)
        )

        # Build response
        resp = CreateRoleResponse(role=self.role_adapter.convert_to_dto(action_result.data))
        return APIResponse.build(status_code=HTTPStatus.CREATED, response_model=resp)

    @auth_required_for_method
    @api_handler
    async def search_roles(
        self,
        body: BodyParam[SearchRolesRequest],
        processors_ctx: ProcessorsCtx,
    ) -> APIResponse:
        """Search roles with filters, orders, and pagination."""
        processors = processors_ctx.processors
        me = current_user()
        if me is None or not me.is_superadmin:
            raise NotEnoughPermission("Only superadmin can search roles.")

        # Build querier using adapter
        querier = self.role_adapter.build_querier(body.parsed)

        # Call service action
        action_result = await processors.permission_controller.search_roles.wait_for_complete(
            SearchRolesAction(querier=querier)
        )

        # Build response
        resp = SearchRolesResponse(
            roles=[self.role_adapter.convert_to_dto(role) for role in action_result.result.items],
            pagination=PaginationInfo(
                total=action_result.result.total_count,
                offset=body.parsed.offset,
                limit=body.parsed.limit,
            ),
        )
        return APIResponse.build(status_code=HTTPStatus.OK, response_model=resp)

    @auth_required_for_method
    @api_handler
    async def get_role(
        self,
        path: PathParam[GetRolePathParam],
        processors_ctx: ProcessorsCtx,
    ) -> APIResponse:
        """Get a specific role with details."""
        processors = processors_ctx.processors
        me = current_user()
        if me is None or not me.is_superadmin:
            raise NotEnoughPermission("Only superadmin can get role details.")

        # Call service action
        action_result = await processors.permission_controller.get_role_detail.wait_for_complete(
            GetRoleDetailAction(role_id=path.parsed.role_id)
        )

        # Build response
        resp = GetRoleResponse(role=self.role_adapter.convert_to_dto(action_result.role))
        return APIResponse.build(status_code=HTTPStatus.OK, response_model=resp)

    @auth_required_for_method
    @api_handler
    async def update_role(
        self,
        path: PathParam[UpdateRolePathParam],
        body: BodyParam[UpdateRoleRequest],
        processors_ctx: ProcessorsCtx,
    ) -> APIResponse:
        """Update an existing role."""
        processors = processors_ctx.processors
        me = current_user()
        if me is None or not me.is_superadmin:
            raise NotEnoughPermission("Only superadmin can update roles.")

        # Build updater using adapter
        role_id = path.parsed.role_id
        updater = self.role_adapter.build_updater(body.parsed, role_id)

        # Call service action
        action_result = await processors.permission_controller.update_role.wait_for_complete(
            UpdateRoleAction(updater=updater)
        )

        # Build response
        resp = UpdateRoleResponse(role=self.role_adapter.convert_to_dto(action_result.data))
        return APIResponse.build(status_code=HTTPStatus.OK, response_model=resp)

    @auth_required_for_method
    @api_handler
    async def delete_role(
        self,
        body: BodyParam[DeleteRoleRequest],
        processors_ctx: ProcessorsCtx,
    ) -> APIResponse:
        """Delete a role (soft delete)."""
        processors = processors_ctx.processors
        me = current_user()
        if me is None or not me.is_superadmin:
            raise NotEnoughPermission("Only superadmin can delete roles.")

        role_id = body.parsed.role_id

        # Create updater
        updater: Updater[RoleRow] = self.role_adapter.build_deleter(role_id)

        # Call service action
        await processors.permission_controller.delete_role.wait_for_complete(
            DeleteRoleAction(updater=updater)
        )

        # Build response
        resp = DeleteRoleResponse(deleted=True)
        return APIResponse.build(status_code=HTTPStatus.OK, response_model=resp)

    @auth_required_for_method
    @api_handler
    async def purge_role(
        self,
        body: BodyParam[PurgeRoleRequest],
        processors_ctx: ProcessorsCtx,
    ) -> APIResponse:
        """Delete a role (soft delete)."""
        processors = processors_ctx.processors
        me = current_user()
        if me is None or not me.is_superadmin:
            raise NotEnoughPermission("Only superadmin can delete roles.")

        role_id = body.parsed.role_id

        # Create purger
        purger: Purger[RoleRow] = self.role_adapter.build_purger(role_id)

        # Call service action
        await processors.permission_controller.purge_role.wait_for_complete(
            PurgeRoleAction(purger=purger)
        )

        # Build response
        resp = DeleteRoleResponse(deleted=True)
        return APIResponse.build(status_code=HTTPStatus.OK, response_model=resp)

    # Role Assignment Endpoints

    @auth_required_for_method
    @api_handler
    async def assign_role(
        self,
        body: BodyParam[AssignRoleRequest],
        processors_ctx: ProcessorsCtx,
    ) -> APIResponse:
        """Assign a role to a user."""
        processors = processors_ctx.processors
        me = current_user()
        if me is None or not me.is_superadmin:
            raise NotEnoughPermission("Only superadmin can assign roles.")

        # Create assignment input
        input_data = UserRoleAssignmentInput(
            user_id=body.parsed.user_id,
            role_id=body.parsed.role_id,
            granted_by=body.parsed.granted_by or me.user_id,
        )

        # Call service action
        action_result = await processors.permission_controller.assign_role.wait_for_complete(
            AssignRoleAction(input=input_data)
        )

        # Build response
        resp = AssignRoleResponse(
            user_id=action_result.data.user_id,
            role_id=action_result.data.role_id,
            granted_by=action_result.data.granted_by,
        )
        return APIResponse.build(status_code=HTTPStatus.CREATED, response_model=resp)

    @auth_required_for_method
    @api_handler
    async def revoke_role(
        self,
        body: BodyParam[RevokeRoleRequest],
        processors_ctx: ProcessorsCtx,
    ) -> APIResponse:
        """Revoke a role from a user."""
        processors = processors_ctx.processors
        me = current_user()
        if me is None or not me.is_superadmin:
            raise NotEnoughPermission("Only superadmin can revoke roles.")

        # Create revocation input
        input_data = UserRoleRevocationInput(
            user_id=body.parsed.user_id,
            role_id=body.parsed.role_id,
        )

        # Call service action
        action_result = await processors.permission_controller.revoke_role.wait_for_complete(
            RevokeRoleAction(input=input_data)
        )

        # Build response
        resp = RevokeRoleResponse(
            user_id=action_result.data.user_id,
            role_id=action_result.data.role_id,
        )
        return APIResponse.build(status_code=HTTPStatus.OK, response_model=resp)

    @auth_required_for_method
    @api_handler
    async def search_assigned_users(
        self,
        path: PathParam[SearchUsersAssignedToRolePathParam],
        body: BodyParam[SearchUsersAssignedToRoleRequest],
        processors_ctx: ProcessorsCtx,
    ) -> APIResponse:
        """Search users assigned to a specific role with filters and pagination."""
        processors = processors_ctx.processors
        me = current_user()
        if me is None or not me.is_superadmin:
            raise NotEnoughPermission("Only superadmin can search assigned users.")

        # Build querier using adapter (includes role_id as filter condition)
        querier = self.assigned_user_adapter.build_querier(path.parsed, body.parsed)

        # Call service action
        action_result = (
            await processors.permission_controller.search_users_assigned_to_role.wait_for_complete(
                SearchUsersAssignedToRoleAction(querier=querier)
            )
        )

        # Build response
        resp = SearchUsersAssignedToRoleResponse(
            users=[
                self.assigned_user_adapter.convert_to_dto(user)
                for user in action_result.result.items
            ],
            pagination=PaginationInfo(
                total=action_result.result.total_count,
                offset=body.parsed.offset,
                limit=body.parsed.limit,
            ),
        )
        return APIResponse.build(status_code=HTTPStatus.OK, response_model=resp)

    # Scope Management Endpoints

    @auth_required_for_method
    @api_handler
    async def get_scope_types(
        self,
        processors_ctx: ProcessorsCtx,
    ) -> APIResponse:
        """Get available scope types for role configuration."""
        processors = processors_ctx.processors
        me = current_user()
        # TODO: Replace with RBAC-based permission check after migration is complete.
        if me is None or not me.is_superadmin:
            raise NotEnoughPermission("Only superadmin can access scope types.")

        # Call service action
        action_result = await processors.permission_controller.get_scope_types.wait_for_complete(
            GetScopeTypesAction()
        )

        # Build response
        resp = GetScopeTypesResponse(items=action_result.scope_types)
        return APIResponse.build(status_code=HTTPStatus.OK, response_model=resp)

    @auth_required_for_method
    @api_handler
    async def search_scopes(
        self,
        path: PathParam[SearchScopesPathParam],
        body: BodyParam[SearchScopesRequest],
        processors_ctx: ProcessorsCtx,
    ) -> APIResponse:
        """Search scopes for a specific scope type with filters and pagination."""
        processors = processors_ctx.processors
        me = current_user()
        # TODO: Replace with RBAC-based permission check after migration is complete.
        if me is None or not me.is_superadmin:
            raise NotEnoughPermission("Only superadmin can search scopes.")

        # Build querier and action
        scope_type = path.parsed.scope_type
        querier = self._scope_adapter.build_querier(scope_type, body.parsed)
        action = SearchScopesAction(scope_type=scope_type, querier=querier)

        # Call service action
        action_result = await processors.permission_controller.search_scopes.wait_for_complete(
            action
        )

        # Build response using adapter
        resp = SearchScopesResponse(
            items=[self._scope_adapter.convert_to_dto(item) for item in action_result.result.items],
            pagination=PaginationInfo(
                total=action_result.result.total_count,
                offset=body.parsed.offset,
                limit=body.parsed.limit,
            ),
        )
        return APIResponse.build(status_code=HTTPStatus.OK, response_model=resp)

    # Entity Management Endpoints

    @auth_required_for_method
    @api_handler
    async def get_entity_types(
        self,
        processors_ctx: ProcessorsCtx,
    ) -> APIResponse:
        """Get available entity types for role configuration."""
        processors = processors_ctx.processors
        me = current_user()
        # TODO: Replace with RBAC-based permission check after migration is complete.
        if me is None or not me.is_superadmin:
            raise NotEnoughPermission("Only superadmin can access entity types.")

        # Call service action
        action_result = await processors.permission_controller.get_entity_types.wait_for_complete(
            GetEntityTypesAction()
        )

        # Build response
        resp = GetEntityTypesResponse(items=action_result.entity_types)
        return APIResponse.build(status_code=HTTPStatus.OK, response_model=resp)

    @auth_required_for_method
    @api_handler
    async def search_entities(
        self,
        path: PathParam[SearchEntitiesPathParam],
        body: BodyParam[SearchEntitiesRequest],
        processors_ctx: ProcessorsCtx,
    ) -> APIResponse:
        """Search entities within a scope by entity type with filters and pagination."""
        processors = processors_ctx.processors
        me = current_user()
        # TODO: Replace with RBAC-based permission check after migration is complete.
        if me is None or not me.is_superadmin:
            raise NotEnoughPermission("Only superadmin can search entities.")

        # Build querier with scope conditions and create action
        querier = self._entity_adapter.build_querier(
            scope_type=path.parsed.scope_type,
            scope_id=path.parsed.scope_id,
            entity_type=path.parsed.entity_type,
            request=body.parsed,
        )
        action = SearchEntitiesAction(querier=querier)

        # Call service action
        action_result = await processors.permission_controller.search_entities.wait_for_complete(
            action
        )

        # Build response using adapter
        resp = SearchEntitiesResponse(
            items=[
                self._entity_adapter.convert_to_dto(item) for item in action_result.result.items
            ],
            pagination=PaginationInfo(
                total=action_result.result.total_count,
                offset=body.parsed.offset,
                limit=body.parsed.limit,
            ),
        )
        return APIResponse.build(status_code=HTTPStatus.OK, response_model=resp)


def create_app(
    default_cors_options: CORSOptions,
) -> tuple[web.Application, Iterable[WebMiddleware]]:
    """Create aiohttp application for RBAC API endpoints."""
    app = web.Application()
    app["api_versions"] = (4, 5)
    app["prefix"] = "v2.0"

    cors = aiohttp_cors.setup(app, defaults=default_cors_options)
    api_handler = RBACAPIHandler()

    # Role management routes
    cors.add(app.router.add_route("POST", "/admin/rbac/roles", api_handler.create_role))
    cors.add(app.router.add_route("POST", "/admin/rbac/roles/search", api_handler.search_roles))
    cors.add(app.router.add_route("GET", "/admin/rbac/roles/{role_id}", api_handler.get_role))
    cors.add(app.router.add_route("PATCH", "/admin/rbac/roles/{role_id}", api_handler.update_role))
    cors.add(app.router.add_route("POST", "/admin/rbac/roles/delete", api_handler.delete_role))
    cors.add(app.router.add_route("POST", "/admin/rbac/roles/purge", api_handler.purge_role))

    # Role assignment routes
    cors.add(app.router.add_route("POST", "/admin/rbac/roles/assign", api_handler.assign_role))
    cors.add(app.router.add_route("POST", "/admin/rbac/roles/revoke", api_handler.revoke_role))
    cors.add(
        app.router.add_route(
            "POST",
            "/admin/rbac/roles/{role_id}/assigned-users/search",
            api_handler.search_assigned_users,
        )
    )

    # Scope routes
    cors.add(app.router.add_route("GET", "/admin/rbac/scope-types", api_handler.get_scope_types))
    cors.add(
        app.router.add_route(
            "POST",
            "/admin/rbac/scopes/{scope_type}/search",
            api_handler.search_scopes,
        )
    )

    # Entity routes
    cors.add(app.router.add_route("GET", "/admin/rbac/entity-types", api_handler.get_entity_types))
    cors.add(
        app.router.add_route(
            "POST",
            "/admin/rbac/scopes/{scope_type}/{scope_id}/entities/{entity_type}/search",
            api_handler.search_entities,
        )
    )

    return app, []



"""User repository layer.

Re-exports public APIs from submodules.
"""

from .options import (
    UserConditions,
    UserOrders,
)
from .repository import (
    UserRepository,
)
from .types import (
    DomainUserSearchScope,
    ProjectUserSearchScope,
)

__all__ = (
    "DomainUserSearchScope",
    "ProjectUserSearchScope",
    "UserConditions",
    "UserOrders",
    "UserRepository",
)



from __future__ import annotations

from collections.abc import AsyncGenerator
from uuid import UUID

import strawberry
from strawberry import ID, Info

from ai.backend.common.data.storage.registries.types import ModelSortKey
from ai.backend.manager.api.gql.base import (
    encode_cursor,
)
from ai.backend.manager.api.gql.types import StrawberryGQLContext
from ai.backend.manager.api.gql.utils import dedent_strip
from ai.backend.manager.data.artifact.types import ArtifactAvailability
from ai.backend.manager.defs import ARTIFACT_MAX_SCAN_LIMIT
from ai.backend.manager.errors.artifact import (
    ArtifactImportDelegationError,
    ArtifactScanLimitExceededError,
)
from ai.backend.manager.repositories.artifact.updaters import ArtifactUpdaterSpec
from ai.backend.manager.repositories.base.updater import Updater
from ai.backend.manager.services.artifact.actions.delegate_scan import DelegateScanArtifactsAction
from ai.backend.manager.services.artifact.actions.delete_multi import DeleteArtifactsAction
from ai.backend.manager.services.artifact.actions.restore_multi import RestoreArtifactsAction
from ai.backend.manager.services.artifact.actions.retrieve_model_multi import RetrieveModelsAction
from ai.backend.manager.services.artifact.actions.scan import ScanArtifactsAction
from ai.backend.manager.services.artifact.actions.update import UpdateArtifactAction
from ai.backend.manager.services.artifact_revision.actions.approve import (
    ApproveArtifactRevisionAction,
)
from ai.backend.manager.services.artifact_revision.actions.cancel_import import CancelImportAction
from ai.backend.manager.services.artifact_revision.actions.cleanup import (
    CleanupArtifactRevisionAction,
)
from ai.backend.manager.services.artifact_revision.actions.delegate_import_revision_batch import (
    DelegateImportArtifactRevisionBatchAction,
)
from ai.backend.manager.services.artifact_revision.actions.import_revision import (
    ImportArtifactRevisionAction,
)
from ai.backend.manager.services.artifact_revision.actions.reject import (
    RejectArtifactRevisionAction,
)
from ai.backend.manager.types import TriState

from .fetcher import (
    fetch_artifact,
    fetch_artifact_revision,
    fetch_artifact_revisions,
    fetch_artifacts,
    get_registry_url,
)
from .types import (
    ApproveArtifactInput,
    ApproveArtifactPayload,
    Artifact,
    ArtifactConnection,
    ArtifactFilter,
    ArtifactImportProgressUpdatedPayload,
    ArtifactOrderBy,
    ArtifactRevision,
    ArtifactRevisionConnection,
    ArtifactRevisionEdge,
    ArtifactRevisionFilter,
    ArtifactRevisionImportTask,
    ArtifactRevisionOrderBy,
    ArtifactStatusChangedInput,
    ArtifactStatusChangedPayload,
    CancelArtifactInput,
    CancelImportArtifactPayload,
    CleanupArtifactRevisionsInput,
    CleanupArtifactRevisionsPayload,
    DelegateImportArtifactsInput,
    DelegateImportArtifactsPayload,
    DelegateScanArtifactsInput,
    DelegateScanArtifactsPayload,
    DeleteArtifactsInput,
    DeleteArtifactsPayload,
    ImportArtifactsInput,
    ImportArtifactsOptionsGQL,
    ImportArtifactsPayload,
    RejectArtifactInput,
    RejectArtifactPayload,
    RestoreArtifactsInput,
    RestoreArtifactsPayload,
    ScanArtifactModelsInput,
    ScanArtifactModelsPayload,
    ScanArtifactsInput,
    ScanArtifactsPayload,
    UpdateArtifactInput,
    UpdateArtifactPayload,
)


# Query Fields
@strawberry.field(  # type: ignore[misc]
    description=dedent_strip("""
    Added in 25.14.0.

    Query artifacts with optional filtering, ordering, and pagination.

    Returns artifacts that are available in the system, discovered through scanning
    external registries like HuggingFace or Reservoir. By default, only shows
    ALIVE (non-deleted) artifacts.

    Use filters to narrow down results by type, name, registry, or availability.
    Supports cursor-based pagination for efficient browsing of large datasets.
    """)
)
async def artifacts(
    info: Info[StrawberryGQLContext],
    filter: ArtifactFilter | None = None,
    order_by: list[ArtifactOrderBy] | None = None,
    before: str | None = None,
    after: str | None = None,
    first: int | None = None,
    last: int | None = None,
    limit: int | None = None,
    offset: int | None = None,
) -> ArtifactConnection | None:
    if filter is None:
        filter = ArtifactFilter(availability=[ArtifactAvailability.ALIVE])
    return await fetch_artifacts(
        info,
        filter,
        order_by,
        before,
        after,
        first,
        last,
        limit,
        offset,
    )


@strawberry.field(  # type: ignore[misc]
    description=dedent_strip("""
    Added in 25.14.0.

    Retrieve a specific artifact by its ID.

    Returns detailed information about the artifact including its metadata,
    registry information, and availability status.
    """)
)
async def artifact(id: ID, info: Info[StrawberryGQLContext]) -> Artifact | None:
    return await fetch_artifact(info, UUID(id))


@strawberry.field(  # type: ignore[misc]
    description=dedent_strip("""
    Added in 25.14.0.

    Query artifact revisions with optional filtering, ordering, and pagination.

    Returns specific versions/revisions of artifacts. Each revision represents
    a specific version of an artifact with its own status, file data, and metadata.

    Use filters to find revisions by status, version, or artifact ID.
    Supports cursor-based pagination for efficient browsing.
    """)
)
async def artifact_revisions(
    info: Info[StrawberryGQLContext],
    filter: ArtifactRevisionFilter | None = None,
    order_by: list[ArtifactRevisionOrderBy] | None = None,
    before: str | None = None,
    after: str | None = None,
    first: int | None = None,
    last: int | None = None,
    limit: int | None = None,
    offset: int | None = None,
) -> ArtifactRevisionConnection | None:
    return await fetch_artifact_revisions(
        info,
        filter,
        order_by,
        before,
        after,
        first,
        last,
        limit,
        offset,
    )


@strawberry.field(  # type: ignore[misc]
    description=dedent_strip("""
    Added in 25.14.0.

    Retrieve a specific artifact revision by its ID.

    Returns detailed information about the revision including its status,
    version, file size, and README content.
    """)
)
async def artifact_revision(id: ID, info: Info[StrawberryGQLContext]) -> ArtifactRevision | None:
    return await fetch_artifact_revision(info, UUID(id))


@strawberry.mutation(  # type: ignore[misc]
    description=dedent_strip("""
    Added in 25.14.0.

    Scan external registries to discover available artifacts.

    Searches HuggingFace or Reservoir registries for artifacts matching the specified
    criteria and registers them in the system with SCANNED status. The artifacts
    become available for import but are not downloaded until explicitly imported.

    This is the first step in the artifact workflow: Scan → Import → Use.
    """)
)
async def scan_artifacts(
    input: ScanArtifactsInput, info: Info[StrawberryGQLContext]
) -> ScanArtifactsPayload:
    if input.limit > ARTIFACT_MAX_SCAN_LIMIT:
        raise ArtifactScanLimitExceededError(f"Limit cannot exceed {ARTIFACT_MAX_SCAN_LIMIT}")

    action_result = await info.context.processors.artifact.scan.wait_for_complete(
        ScanArtifactsAction(
            artifact_type=input.artifact_type,
            registry_id=UUID(input.registry_id) if input.registry_id else None,
            limit=input.limit,
            # TODO: Move this huggingface_registries config if needed
            order=ModelSortKey.DOWNLOADS,
            search=input.search,
        )
    )

    data_loaders = info.context.data_loaders

    artifacts = []
    for item in action_result.result:
        registry_url = await get_registry_url(data_loaders, item.registry_id, item.registry_type)
        source_url = await get_registry_url(
            data_loaders, item.source_registry_id, item.source_registry_type
        )
        artifacts.append(Artifact.from_dataclass(item, registry_url, source_url))
    return ScanArtifactsPayload(artifacts=artifacts)


# Mutations
@strawberry.mutation(  # type: ignore[misc]
    description=dedent_strip("""
    Added in 25.14.0.

    Import scanned artifact revisions from external registries.

    Downloads the actual files for the specified artifact revisions, transitioning
    them from SCANNED → PULLING → VERIFYING → AVAILABLE status.

    Returns background tasks that can be monitored for import progress.
    Once AVAILABLE, artifacts can be used by users in their sessions.
    """)
)
async def import_artifacts(
    input: ImportArtifactsInput, info: Info[StrawberryGQLContext]
) -> ImportArtifactsPayload:
    imported_artifacts = []
    tasks = []
    vfolder_id = UUID(input.vfolder_id) if input.vfolder_id else None
    # When using VFolderStorage (vfolder_id provided), store at root path
    storage_prefix = "/" if vfolder_id else None
    options = input.options or ImportArtifactsOptionsGQL()
    force = options.force
    for revision_id in input.artifact_revision_ids:
        action_result = (
            await info.context.processors.artifact_revision.import_revision.wait_for_complete(
                ImportArtifactRevisionAction(
                    artifact_revision_id=UUID(revision_id),
                    vfolder_id=vfolder_id,
                    storage_prefix=storage_prefix,
                    force=force,
                )
            )
        )
        artifact_revision = ArtifactRevision.from_dataclass(action_result.result)
        imported_artifacts.append(artifact_revision)
        tasks.append(
            ArtifactRevisionImportTask(
                task_id=ID(str(action_result.task_id)),
                artifact_revision=artifact_revision,
            )
        )

    edges = [
        ArtifactRevisionEdge(node=artifact, cursor=encode_cursor(artifact.id))
        for artifact in imported_artifacts
    ]

    artifacts_connection = ArtifactRevisionConnection(
        count=len(imported_artifacts),
        edges=edges,
        page_info=strawberry.relay.PageInfo(
            has_next_page=False,
            has_previous_page=False,
            start_cursor=edges[0].cursor if edges else None,
            end_cursor=edges[-1].cursor if edges else None,
        ),
    )

    return ImportArtifactsPayload(artifact_revisions=artifacts_connection, tasks=tasks)


@strawberry.mutation(  # type: ignore[misc]
    description=dedent_strip("""
    Added in 25.15.0.

    Triggers artifact scanning on a remote reservoir registry.

    This mutation instructs a reservoir-type registry to initiate a scan of artifacts
    from its associated remote reservoir registry source. The scan process will discover and
    catalog artifacts available in the remote reservoir, making them accessible
    through the local reservoir registry.

    Requirements:
    - The delegator registry must be of type 'reservoir'
    - The delegator reservoir registry must have a valid remote registry configuration
""")
)
async def delegate_scan_artifacts(
    input: DelegateScanArtifactsInput, info: Info[StrawberryGQLContext]
) -> DelegateScanArtifactsPayload:
    if input.limit > ARTIFACT_MAX_SCAN_LIMIT:
        raise ArtifactScanLimitExceededError(f"Limit cannot exceed {ARTIFACT_MAX_SCAN_LIMIT}")

    delegator_reservoir_id = (
        UUID(input.delegator_reservoir_id) if input.delegator_reservoir_id else None
    )

    action_result = await info.context.processors.artifact.delegate_scan.wait_for_complete(
        DelegateScanArtifactsAction(
            delegator_reservoir_id=delegator_reservoir_id,
            delegatee_target=input.delegatee_target.to_dataclass()
            if input.delegatee_target
            else None,
            artifact_type=input.artifact_type,
            limit=input.limit,
            order=ModelSortKey.DOWNLOADS,
            search=input.search,
        )
    )

    data_loaders = info.context.data_loaders

    artifacts = []
    for item in action_result.result:
        registry_url = await get_registry_url(data_loaders, item.registry_id, item.registry_type)
        source_url = await get_registry_url(
            data_loaders, item.source_registry_id, item.source_registry_type
        )
        artifacts.append(Artifact.from_dataclass(item, registry_url, source_url))
    return DelegateScanArtifactsPayload(artifacts=artifacts)


@strawberry.mutation(  # type: ignore[misc]
    description=dedent_strip("""
    Added in 25.15.0.

    Trigger import of artifact revisions from a remote reservoir registry.

    This mutation instructs a reservoir-type registry to import specific artifact revisions
    that were previously discovered during a scan from its remote registry.
    Note that this operation does not import the artifacts directly into the local registry, but only into the delegator reservoir's storage.

    Requirements:
    - The delegator registry must be of type 'reservoir'
    - The delegator registry must have a valid remote registry configuration
""")
)
async def delegate_import_artifacts(
    input: DelegateImportArtifactsInput, info: Info[StrawberryGQLContext]
) -> DelegateImportArtifactsPayload:
    imported_artifacts = []
    tasks = []

    options = input.options or ImportArtifactsOptionsGQL()
    force = options.force
    action_result = await info.context.processors.artifact_revision.delegate_import_revision_batch.wait_for_complete(
        DelegateImportArtifactRevisionBatchAction(
            delegator_reservoir_id=UUID(input.delegator_reservoir_id)
            if input.delegator_reservoir_id
            else None,
            delegatee_target=input.delegatee_target.to_dataclass()
            if input.delegatee_target
            else None,
            artifact_type=input.artifact_type,
            artifact_revision_ids=[
                UUID(revision_id) for revision_id in input.artifact_revision_ids
            ],
            force=force,
        )
    )
    artifact_revisions = [
        ArtifactRevision.from_dataclass(result) for result in action_result.result
    ]
    imported_artifacts.extend(artifact_revisions)

    if len(artifact_revisions) != len(action_result.task_ids):
        raise ArtifactImportDelegationError(
            "Mismatch between artifact revisions and task IDs returned"
        )

    for task_uuid, artifact_revision in zip(
        action_result.task_ids, artifact_revisions, strict=True
    ):
        task_id = ID(str(task_uuid)) if task_uuid is not None else None
        tasks.append(
            ArtifactRevisionImportTask(
                task_id=task_id,
                artifact_revision=artifact_revision,
            )
        )

    edges = [
        ArtifactRevisionEdge(node=artifact, cursor=encode_cursor(artifact.id))
        for artifact in imported_artifacts
    ]

    artifacts_connection = ArtifactRevisionConnection(
        count=len(imported_artifacts),
        edges=edges,
        page_info=strawberry.relay.PageInfo(
            has_next_page=False,
            has_previous_page=False,
            start_cursor=edges[0].cursor if edges else None,
            end_cursor=edges[-1].cursor if edges else None,
        ),
    )

    return DelegateImportArtifactsPayload(artifact_revisions=artifacts_connection, tasks=tasks)


@strawberry.mutation(  # type: ignore[misc]
    description=dedent_strip("""
    Added in 25.14.0.

    Update artifact metadata properties.

    Modifies artifact metadata such as readonly status and description.
    This operation does not affect the actual artifact files or revisions.
    """)
)
async def update_artifact(
    input: UpdateArtifactInput, info: Info[StrawberryGQLContext]
) -> UpdateArtifactPayload:
    action_result = await info.context.processors.artifact.update.wait_for_complete(
        UpdateArtifactAction(
            updater=Updater(
                spec=ArtifactUpdaterSpec(
                    readonly=TriState.from_graphql(input.readonly),
                    description=TriState.from_graphql(input.description),
                ),
                pk_value=UUID(input.artifact_id),
            ),
        )
    )

    artifact = action_result.result
    data_loaders = info.context.data_loaders

    registry_url = await get_registry_url(
        data_loaders, artifact.registry_id, artifact.registry_type
    )
    source_url = await get_registry_url(
        data_loaders, artifact.source_registry_id, artifact.source_registry_type
    )

    return UpdateArtifactPayload(
        artifact=Artifact.from_dataclass(artifact, registry_url=registry_url, source_url=source_url)
    )


@strawberry.mutation(  # type: ignore[misc]
    description=dedent_strip("""
    Added in 25.14.0.

    Clean up stored artifact revision data to free storage space.

    Removes the downloaded files for the specified artifact revisions and
    transitions them back to SCANNED status. The metadata remains, allowing
    the artifacts to be re-imported later if needed.

    Use this operation to manage storage usage by removing unused artifacts.
    """)
)
async def cleanup_artifact_revisions(
    input: CleanupArtifactRevisionsInput, info: Info[StrawberryGQLContext]
) -> CleanupArtifactRevisionsPayload:
    cleaned_artifact_revisions: list[ArtifactRevision] = []
    # TODO: Refactor with asyncio.gather()
    for artifact_revision_id in input.artifact_revision_ids:
        action_result = await info.context.processors.artifact_revision.cleanup.wait_for_complete(
            CleanupArtifactRevisionAction(
                artifact_revision_id=UUID(artifact_revision_id),
            )
        )
        cleaned_artifact_revisions.append(ArtifactRevision.from_dataclass(action_result.result))

    edges = [
        ArtifactRevisionEdge(node=revision, cursor=encode_cursor(revision.id))
        for revision in cleaned_artifact_revisions
    ]

    artifacts_connection = ArtifactRevisionConnection(
        count=len(cleaned_artifact_revisions),
        edges=edges,
        page_info=strawberry.relay.PageInfo(
            has_next_page=False,
            has_previous_page=False,
            start_cursor=edges[0].cursor if edges else None,
            end_cursor=edges[-1].cursor if edges else None,
        ),
    )

    return CleanupArtifactRevisionsPayload(artifact_revisions=artifacts_connection)


@strawberry.mutation(  # type: ignore[misc]
    description=dedent_strip("""
    Added in 25.15.0.

    Soft-delete artifacts from the system.

    Marks artifacts as deleted without permanently removing them.
    Deleted artifacts can be restored using the restore_artifacts mutation.
    """)
)
async def delete_artifacts(
    input: DeleteArtifactsInput, info: Info[StrawberryGQLContext]
) -> DeleteArtifactsPayload:
    action_result = await info.context.processors.artifact.delete_artifacts.wait_for_complete(
        DeleteArtifactsAction(
            artifact_ids=[UUID(id) for id in input.artifact_ids],
        )
    )

    data_loaders = info.context.data_loaders

    artifacts = []
    for item in action_result.artifacts:
        registry_url = await get_registry_url(data_loaders, item.registry_id, item.registry_type)
        source_url = await get_registry_url(
            data_loaders, item.source_registry_id, item.source_registry_type
        )
        artifacts.append(Artifact.from_dataclass(item, registry_url, source_url))

    return DeleteArtifactsPayload(artifacts=artifacts)


@strawberry.mutation(  # type: ignore[misc]
    description=dedent_strip("""
    Added in 25.15.0.

    Restore previously deleted artifacts.

    Reverses the soft-delete operation, making the artifacts available again
    for use in the system.
    """)
)
async def restore_artifacts(
    input: RestoreArtifactsInput, info: Info[StrawberryGQLContext]
) -> RestoreArtifactsPayload:
    action_result = await info.context.processors.artifact.restore_artifacts.wait_for_complete(
        RestoreArtifactsAction(
            artifact_ids=[UUID(id) for id in input.artifact_ids],
        )
    )

    data_loaders = info.context.data_loaders

    artifacts = []
    for item in action_result.artifacts:
        registry_url = await get_registry_url(data_loaders, item.registry_id, item.registry_type)
        source_url = await get_registry_url(
            data_loaders, item.source_registry_id, item.source_registry_type
        )
        artifacts.append(Artifact.from_dataclass(item, registry_url, source_url))

    return RestoreArtifactsPayload(artifacts=artifacts)


@strawberry.mutation(  # type: ignore[misc]
    description=dedent_strip("""
    Added in 25.14.0.

    Cancel an in-progress artifact import operation.

    Stops the download process for the specified artifact revision and
    reverts its status back to SCANNED. The partially downloaded data is cleaned up.
    """)
)
async def cancel_import_artifact(
    input: CancelArtifactInput, info: Info[StrawberryGQLContext]
) -> CancelImportArtifactPayload:
    # TODO: Cancel actual import bgtask
    action_result = await info.context.processors.artifact_revision.cancel_import.wait_for_complete(
        CancelImportAction(
            artifact_revision_id=UUID(input.artifact_revision_id),
        )
    )
    return CancelImportArtifactPayload(
        artifact_revision=ArtifactRevision.from_dataclass(action_result.result)
    )


# TODO: Make this available when only having super-admin privileges
@strawberry.mutation(  # type: ignore[misc]
    description=dedent_strip("""
    Added in 25.14.0.

    Approve an artifact revision for general use.

    Admin-only operation to approve artifact revisions, typically used
    in environments with approval workflows for artifact deployment.
    """)
)
async def approve_artifact_revision(
    input: ApproveArtifactInput, info: Info[StrawberryGQLContext]
) -> ApproveArtifactPayload:
    action_result = await info.context.processors.artifact_revision.approve.wait_for_complete(
        ApproveArtifactRevisionAction(
            artifact_revision_id=UUID(input.artifact_revision_id),
        )
    )

    return ApproveArtifactPayload(
        artifact_revision=ArtifactRevision.from_dataclass(action_result.result)
    )


@strawberry.mutation(  # type: ignore[misc]
    description=dedent_strip("""
    Added in 25.14.0.

    Reject an artifact revision, preventing its use.

    Admin-only operation to reject artifact revisions, typically used
    in environments with approval workflows for artifact deployment.
    """)
)
async def reject_artifact_revision(
    input: RejectArtifactInput, info: Info[StrawberryGQLContext]
) -> RejectArtifactPayload:
    action_result = await info.context.processors.artifact_revision.reject.wait_for_complete(
        RejectArtifactRevisionAction(
            artifact_revision_id=UUID(input.artifact_revision_id),
        )
    )

    return RejectArtifactPayload(
        artifact_revision=ArtifactRevision.from_dataclass(action_result.result)
    )


@strawberry.mutation(  # type: ignore[misc]
    description=dedent_strip("""
    Added in 25.14.0.

    Perform detailed scanning of specific models.

    Unlike the general scan_artifacts operation, this performs immediate detailed
    scanning of specified models including README content and file sizes.
    Returns artifact revisions with complete metadata ready for use.
    """)
)
async def scan_artifact_models(
    input: ScanArtifactModelsInput, info: Info[StrawberryGQLContext]
) -> ScanArtifactModelsPayload:
    action_result = await info.context.processors.artifact.retrieve_models.wait_for_complete(
        RetrieveModelsAction(
            models=[m.to_dataclass() for m in input.models], registry_id=input.registry_id
        )
    )

    edges = []
    for data in action_result.result:
        edges.extend([
            ArtifactRevisionEdge(
                node=ArtifactRevision.from_dataclass(revision),
                cursor=encode_cursor(revision.id),
            )
            for revision in data.revisions
        ])

    artifacts_connection = ArtifactRevisionConnection(
        count=len(edges),
        edges=edges,
        page_info=strawberry.relay.PageInfo(
            has_next_page=False,
            has_previous_page=False,
            start_cursor=edges[0].cursor if edges else None,
            end_cursor=edges[-1].cursor if edges else None,
        ),
    )

    return ScanArtifactModelsPayload(artifact_revision=artifacts_connection)


# Subscriptions
@strawberry.subscription(  # type: ignore[misc]
    description=dedent_strip("""
    Added in 25.14.0.

    Subscribe to real-time artifact status change notifications.

    Receives updates when artifact revision statuses change during import,
    cleanup, or other operations. Useful for building reactive UIs that
    show live progress of artifact operations.
    """)
)
async def artifact_status_changed(
    input: ArtifactStatusChangedInput,
) -> AsyncGenerator[ArtifactStatusChangedPayload, None]:
    # TODO: Implement actual subscription logic using pub/sub system
    raise NotImplementedError("Subscription not implemented")
    yield  # type: ignore[unreachable]  # Makes this an async generator


@strawberry.subscription(  # type: ignore[misc]
    description=dedent_strip("""
    Added in 25.14.0.

    Subscribe to real-time artifact import progress updates.

    Receives progress notifications during artifact import operations,
    including percentage completed and current status. Useful for displaying
    progress bars and real-time import status to users.
    """)
)
async def artifact_import_progress_updated(
    artifact_revision_id: ID,
) -> AsyncGenerator[ArtifactImportProgressUpdatedPayload, None]:
    # TODO: Implement actual subscription logic using pub/sub system
    raise NotImplementedError("Subscription not implemented")
    yield  # type: ignore[unreachable]  # Makes this an async generator



"""fix typo in column name

Revision ID: ffbebdf3c2b9
Revises: 38741c85f228
Create Date: 2024-05-20 11:30:59.083088

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "ffbebdf3c2b9"
down_revision = "38741c85f228"
branch_labels = None
depends_on = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    conn = op.get_bind()
    op.add_column(
        "endpoints",
        sa.Column(
            "model_mount_destination",
            sa.String(length=1024),
            server_default="/models",
            nullable=False,
        ),
    )
    conn.exec_driver_sql("UPDATE endpoints SET model_mount_destination = model_mount_destiation")
    op.drop_column("endpoints", "model_mount_destiation")
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    conn = op.get_bind()
    op.add_column(
        "endpoints",
        sa.Column(
            "model_mount_destiation",
            sa.VARCHAR(length=1024),
            server_default=sa.text("'/models'::character varying"),
            autoincrement=False,
            nullable=False,
        ),
    )
    conn.exec_driver_sql("UPDATE endpoints SET model_mount_destiation = model_mount_destination")
    op.drop_column("endpoints", "model_mount_destination")
    # ### end Alembic commands ###



from __future__ import annotations

import uuid
from collections.abc import Collection
from datetime import datetime
from typing import cast

import sqlalchemy as sa

from ai.backend.common.types import KernelId, SessionId
from ai.backend.manager.api.gql.base import StringMatchSpec, UUIDEqualMatchSpec, UUIDInMatchSpec
from ai.backend.manager.data.deployment.types import RouteStatus
from ai.backend.manager.data.kernel.types import KernelSchedulingPhase
from ai.backend.manager.data.session.types import SchedulingResult, SessionStatus
from ai.backend.manager.models.scheduling_history import (
    DeploymentHistoryRow,
    KernelSchedulingHistoryRow,
    RouteHistoryRow,
    SessionSchedulingHistoryRow,
)
from ai.backend.manager.repositories.base import QueryCondition, QueryOrder

# ========== Session Scheduling History ==========


class SessionSchedulingHistoryConditions:
    """Query conditions for session scheduling history."""

    # UUID filter conditions for history id
    @staticmethod
    def by_id_filter(spec: UUIDEqualMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            if spec.negated:
                return SessionSchedulingHistoryRow.id != spec.value
            return SessionSchedulingHistoryRow.id == spec.value

        return inner

    @staticmethod
    def by_id_in(spec: UUIDInMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            if spec.negated:
                return SessionSchedulingHistoryRow.id.notin_(spec.values)
            return SessionSchedulingHistoryRow.id.in_(spec.values)

        return inner

    @staticmethod
    def by_ids(ids: Collection[uuid.UUID]) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return SessionSchedulingHistoryRow.id.in_(ids)

        return inner

    @staticmethod
    def by_session_id(session_id: SessionId) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return SessionSchedulingHistoryRow.session_id == session_id

        return inner

    @staticmethod
    def by_result(result: SchedulingResult) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return SessionSchedulingHistoryRow.result == str(result)

        return inner

    @staticmethod
    def by_results(results: list[SchedulingResult]) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return SessionSchedulingHistoryRow.result.in_([str(r) for r in results])

        return inner

    @staticmethod
    def by_from_status(status: SessionStatus) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return SessionSchedulingHistoryRow.from_status == str(status)

        return inner

    @staticmethod
    def by_from_statuses(statuses: list[str]) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return SessionSchedulingHistoryRow.from_status.in_(statuses)

        return inner

    @staticmethod
    def by_to_status(status: SessionStatus) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return SessionSchedulingHistoryRow.to_status == str(status)

        return inner

    @staticmethod
    def by_to_statuses(statuses: list[str]) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return SessionSchedulingHistoryRow.to_status.in_(statuses)

        return inner

    @staticmethod
    def by_error_code(error_code: str) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return SessionSchedulingHistoryRow.error_code == error_code

        return inner

    # UUID filter conditions for session_id
    @staticmethod
    def by_session_id_filter(spec: UUIDEqualMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            if spec.negated:
                return SessionSchedulingHistoryRow.session_id != spec.value
            return SessionSchedulingHistoryRow.session_id == spec.value

        return inner

    @staticmethod
    def by_session_id_in(spec: UUIDInMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            if spec.negated:
                return SessionSchedulingHistoryRow.session_id.notin_(spec.values)
            return SessionSchedulingHistoryRow.session_id.in_(spec.values)

        return inner

    # String filter conditions for error_code
    @staticmethod
    def by_error_code_contains(spec: StringMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            col = cast(sa.ColumnElement[str | None], SessionSchedulingHistoryRow.error_code)
            if spec.case_insensitive:
                col = sa.func.lower(col)
                pattern = f"%{spec.value.lower()}%"
            else:
                pattern = f"%{spec.value}%"
            expr = col.like(pattern)
            return ~expr if spec.negated else expr

        return inner

    @staticmethod
    def by_error_code_equals(spec: StringMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            col = cast(sa.ColumnElement[str | None], SessionSchedulingHistoryRow.error_code)
            if spec.case_insensitive:
                col = sa.func.lower(col)
                val = spec.value.lower()
            else:
                val = spec.value
            if spec.negated:
                return col != val
            return col == val

        return inner

    @staticmethod
    def by_error_code_starts_with(spec: StringMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            col = cast(sa.ColumnElement[str | None], SessionSchedulingHistoryRow.error_code)
            if spec.case_insensitive:
                col = sa.func.lower(col)
                pattern = f"{spec.value.lower()}%"
            else:
                pattern = f"{spec.value}%"
            expr = col.like(pattern)
            return ~expr if spec.negated else expr

        return inner

    @staticmethod
    def by_error_code_ends_with(spec: StringMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            col = cast(sa.ColumnElement[str | None], SessionSchedulingHistoryRow.error_code)
            if spec.case_insensitive:
                col = sa.func.lower(col)
                pattern = f"%{spec.value.lower()}"
            else:
                pattern = f"%{spec.value}"
            expr = col.like(pattern)
            return ~expr if spec.negated else expr

        return inner

    # String filter conditions for phase
    @staticmethod
    def by_phase_contains(spec: StringMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            col = cast(sa.ColumnElement[str], SessionSchedulingHistoryRow.phase)
            if spec.case_insensitive:
                col = sa.func.lower(col)
                pattern = f"%{spec.value.lower()}%"
            else:
                pattern = f"%{spec.value}%"
            expr = col.like(pattern)
            return ~expr if spec.negated else expr

        return inner

    @staticmethod
    def by_phase_equals(spec: StringMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            col = cast(sa.ColumnElement[str], SessionSchedulingHistoryRow.phase)
            if spec.case_insensitive:
                col = sa.func.lower(col)
                val = spec.value.lower()
            else:
                val = spec.value
            if spec.negated:
                return col != val
            return col == val

        return inner

    @staticmethod
    def by_phase_starts_with(spec: StringMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            col = cast(sa.ColumnElement[str], SessionSchedulingHistoryRow.phase)
            if spec.case_insensitive:
                col = sa.func.lower(col)
                pattern = f"{spec.value.lower()}%"
            else:
                pattern = f"{spec.value}%"
            expr = col.like(pattern)
            return ~expr if spec.negated else expr

        return inner

    @staticmethod
    def by_phase_ends_with(spec: StringMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            col = cast(sa.ColumnElement[str], SessionSchedulingHistoryRow.phase)
            if spec.case_insensitive:
                col = sa.func.lower(col)
                pattern = f"%{spec.value.lower()}"
            else:
                pattern = f"%{spec.value}"
            expr = col.like(pattern)
            return ~expr if spec.negated else expr

        return inner

    # String filter conditions for message
    @staticmethod
    def by_message_contains(spec: StringMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            col = cast(sa.ColumnElement[str | None], SessionSchedulingHistoryRow.message)
            if spec.case_insensitive:
                col = sa.func.lower(col)
                pattern = f"%{spec.value.lower()}%"
            else:
                pattern = f"%{spec.value}%"
            expr = col.like(pattern)
            return ~expr if spec.negated else expr

        return inner

    @staticmethod
    def by_message_equals(spec: StringMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            col = cast(sa.ColumnElement[str | None], SessionSchedulingHistoryRow.message)
            if spec.case_insensitive:
                col = sa.func.lower(col)
                val = spec.value.lower()
            else:
                val = spec.value
            if spec.negated:
                return col != val
            return col == val

        return inner

    @staticmethod
    def by_message_starts_with(spec: StringMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            col = cast(sa.ColumnElement[str | None], SessionSchedulingHistoryRow.message)
            if spec.case_insensitive:
                col = sa.func.lower(col)
                pattern = f"{spec.value.lower()}%"
            else:
                pattern = f"{spec.value}%"
            expr = col.like(pattern)
            return ~expr if spec.negated else expr

        return inner

    @staticmethod
    def by_message_ends_with(spec: StringMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            col = cast(sa.ColumnElement[str | None], SessionSchedulingHistoryRow.message)
            if spec.case_insensitive:
                col = sa.func.lower(col)
                pattern = f"%{spec.value.lower()}"
            else:
                pattern = f"%{spec.value}"
            expr = col.like(pattern)
            return ~expr if spec.negated else expr

        return inner

    @staticmethod
    def by_cursor_forward(cursor_id: str) -> QueryCondition:
        """Cursor condition for forward pagination (after cursor)."""
        cursor_uuid = uuid.UUID(cursor_id)

        def inner() -> sa.sql.expression.ColumnElement[bool]:
            subquery = (
                sa.select(SessionSchedulingHistoryRow.created_at)
                .where(SessionSchedulingHistoryRow.id == cursor_uuid)
                .scalar_subquery()
            )
            return SessionSchedulingHistoryRow.created_at < subquery

        return inner

    @staticmethod
    def by_cursor_backward(cursor_id: str) -> QueryCondition:
        """Cursor condition for backward pagination (before cursor)."""
        cursor_uuid = uuid.UUID(cursor_id)

        def inner() -> sa.sql.expression.ColumnElement[bool]:
            subquery = (
                sa.select(SessionSchedulingHistoryRow.created_at)
                .where(SessionSchedulingHistoryRow.id == cursor_uuid)
                .scalar_subquery()
            )
            return SessionSchedulingHistoryRow.created_at > subquery

        return inner

    # DateTime filter conditions
    @staticmethod
    def by_created_at_before(dt: datetime) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return SessionSchedulingHistoryRow.created_at < dt

        return inner

    @staticmethod
    def by_created_at_after(dt: datetime) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return SessionSchedulingHistoryRow.created_at > dt

        return inner

    @staticmethod
    def by_created_at_equals(dt: datetime) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return SessionSchedulingHistoryRow.created_at == dt

        return inner

    @staticmethod
    def by_updated_at_before(dt: datetime) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return SessionSchedulingHistoryRow.updated_at < dt

        return inner

    @staticmethod
    def by_updated_at_after(dt: datetime) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return SessionSchedulingHistoryRow.updated_at > dt

        return inner

    @staticmethod
    def by_updated_at_equals(dt: datetime) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return SessionSchedulingHistoryRow.updated_at == dt

        return inner


class SessionSchedulingHistoryOrders:
    """Query orders for session scheduling history."""

    @staticmethod
    def created_at(ascending: bool = True) -> QueryOrder:
        if ascending:
            return SessionSchedulingHistoryRow.created_at.asc()
        return SessionSchedulingHistoryRow.created_at.desc()

    @staticmethod
    def updated_at(ascending: bool = True) -> QueryOrder:
        if ascending:
            return SessionSchedulingHistoryRow.updated_at.asc()
        return SessionSchedulingHistoryRow.updated_at.desc()


# ========== Kernel Scheduling History ==========


class KernelSchedulingHistoryConditions:
    """Query conditions for kernel scheduling history."""

    @staticmethod
    def by_kernel_id(kernel_id: KernelId) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return KernelSchedulingHistoryRow.kernel_id == kernel_id

        return inner

    @staticmethod
    def by_session_id(session_id: SessionId) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return KernelSchedulingHistoryRow.session_id == session_id

        return inner

    @staticmethod
    def by_result(result: SchedulingResult) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return KernelSchedulingHistoryRow.result == str(result)

        return inner

    @staticmethod
    def by_from_phase(phase: KernelSchedulingPhase) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return cast(
                sa.sql.expression.ColumnElement[bool],
                KernelSchedulingHistoryRow.from_phase == str(phase),
            )

        return inner

    @staticmethod
    def by_to_phase(phase: KernelSchedulingPhase) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return cast(
                sa.sql.expression.ColumnElement[bool],
                KernelSchedulingHistoryRow.to_phase == str(phase),
            )

        return inner

    @staticmethod
    def by_error_code(error_code: str) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return KernelSchedulingHistoryRow.error_code == error_code

        return inner

    @staticmethod
    def by_cursor_forward(cursor_id: str) -> QueryCondition:
        """Cursor condition for forward pagination (after cursor)."""
        cursor_uuid = uuid.UUID(cursor_id)

        def inner() -> sa.sql.expression.ColumnElement[bool]:
            subquery = (
                sa.select(KernelSchedulingHistoryRow.created_at)
                .where(KernelSchedulingHistoryRow.id == cursor_uuid)
                .scalar_subquery()
            )
            return KernelSchedulingHistoryRow.created_at < subquery

        return inner

    @staticmethod
    def by_cursor_backward(cursor_id: str) -> QueryCondition:
        """Cursor condition for backward pagination (before cursor)."""
        cursor_uuid = uuid.UUID(cursor_id)

        def inner() -> sa.sql.expression.ColumnElement[bool]:
            subquery = (
                sa.select(KernelSchedulingHistoryRow.created_at)
                .where(KernelSchedulingHistoryRow.id == cursor_uuid)
                .scalar_subquery()
            )
            return KernelSchedulingHistoryRow.created_at > subquery

        return inner


class KernelSchedulingHistoryOrders:
    """Query orders for kernel scheduling history."""

    @staticmethod
    def created_at(ascending: bool = True) -> QueryOrder:
        if ascending:
            return KernelSchedulingHistoryRow.created_at.asc()
        return KernelSchedulingHistoryRow.created_at.desc()

    @staticmethod
    def updated_at(ascending: bool = True) -> QueryOrder:
        if ascending:
            return KernelSchedulingHistoryRow.updated_at.asc()
        return KernelSchedulingHistoryRow.updated_at.desc()


# ========== Deployment History ==========


class DeploymentHistoryConditions:
    """Query conditions for deployment history."""

    # UUID filter conditions for history id
    @staticmethod
    def by_id_filter(spec: UUIDEqualMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            if spec.negated:
                return DeploymentHistoryRow.id != spec.value
            return DeploymentHistoryRow.id == spec.value

        return inner

    @staticmethod
    def by_id_in(spec: UUIDInMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            if spec.negated:
                return DeploymentHistoryRow.id.notin_(spec.values)
            return DeploymentHistoryRow.id.in_(spec.values)

        return inner

    @staticmethod
    def by_ids(ids: Collection[uuid.UUID]) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return DeploymentHistoryRow.id.in_(ids)

        return inner

    @staticmethod
    def by_deployment_id(deployment_id: uuid.UUID) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return DeploymentHistoryRow.deployment_id == deployment_id

        return inner

    @staticmethod
    def by_result(result: SchedulingResult) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return DeploymentHistoryRow.result == str(result)

        return inner

    @staticmethod
    def by_results(results: list[SchedulingResult]) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return DeploymentHistoryRow.result.in_([str(r) for r in results])

        return inner

    @staticmethod
    def by_error_code(error_code: str) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return DeploymentHistoryRow.error_code == error_code

        return inner

    # UUID filter conditions for deployment_id
    @staticmethod
    def by_deployment_id_filter(spec: UUIDEqualMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            if spec.negated:
                return DeploymentHistoryRow.deployment_id != spec.value
            return DeploymentHistoryRow.deployment_id == spec.value

        return inner

    @staticmethod
    def by_deployment_id_in(spec: UUIDInMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            if spec.negated:
                return DeploymentHistoryRow.deployment_id.notin_(spec.values)
            return DeploymentHistoryRow.deployment_id.in_(spec.values)

        return inner

    # String filter conditions for error_code
    @staticmethod
    def by_error_code_contains(spec: StringMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            col = cast(sa.ColumnElement[str | None], DeploymentHistoryRow.error_code)
            if spec.case_insensitive:
                col = sa.func.lower(col)
                pattern = f"%{spec.value.lower()}%"
            else:
                pattern = f"%{spec.value}%"
            expr = col.like(pattern)
            return ~expr if spec.negated else expr

        return inner

    @staticmethod
    def by_error_code_equals(spec: StringMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            col = cast(sa.ColumnElement[str | None], DeploymentHistoryRow.error_code)
            if spec.case_insensitive:
                col = sa.func.lower(col)
                val = spec.value.lower()
            else:
                val = spec.value
            if spec.negated:
                return col != val
            return col == val

        return inner

    @staticmethod
    def by_error_code_starts_with(spec: StringMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            col = cast(sa.ColumnElement[str | None], DeploymentHistoryRow.error_code)
            if spec.case_insensitive:
                col = sa.func.lower(col)
                pattern = f"{spec.value.lower()}%"
            else:
                pattern = f"{spec.value}%"
            expr = col.like(pattern)
            return ~expr if spec.negated else expr

        return inner

    @staticmethod
    def by_error_code_ends_with(spec: StringMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            col = cast(sa.ColumnElement[str | None], DeploymentHistoryRow.error_code)
            if spec.case_insensitive:
                col = sa.func.lower(col)
                pattern = f"%{spec.value.lower()}"
            else:
                pattern = f"%{spec.value}"
            expr = col.like(pattern)
            return ~expr if spec.negated else expr

        return inner

    # Status list filter conditions
    @staticmethod
    def by_from_statuses(statuses: list[str]) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return DeploymentHistoryRow.from_status.in_(statuses)

        return inner

    @staticmethod
    def by_to_statuses(statuses: list[str]) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return DeploymentHistoryRow.to_status.in_(statuses)

        return inner

    # String filter conditions for phase
    @staticmethod
    def by_phase_contains(spec: StringMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            col = cast(sa.ColumnElement[str], DeploymentHistoryRow.phase)
            if spec.case_insensitive:
                col = sa.func.lower(col)
                pattern = f"%{spec.value.lower()}%"
            else:
                pattern = f"%{spec.value}%"
            expr = col.like(pattern)
            return ~expr if spec.negated else expr

        return inner

    @staticmethod
    def by_phase_equals(spec: StringMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            col = cast(sa.ColumnElement[str], DeploymentHistoryRow.phase)
            if spec.case_insensitive:
                col = sa.func.lower(col)
                val = spec.value.lower()
            else:
                val = spec.value
            if spec.negated:
                return col != val
            return col == val

        return inner

    @staticmethod
    def by_phase_starts_with(spec: StringMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            col = cast(sa.ColumnElement[str], DeploymentHistoryRow.phase)
            if spec.case_insensitive:
                col = sa.func.lower(col)
                pattern = f"{spec.value.lower()}%"
            else:
                pattern = f"{spec.value}%"
            expr = col.like(pattern)
            return ~expr if spec.negated else expr

        return inner

    @staticmethod
    def by_phase_ends_with(spec: StringMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            col = cast(sa.ColumnElement[str], DeploymentHistoryRow.phase)
            if spec.case_insensitive:
                col = sa.func.lower(col)
                pattern = f"%{spec.value.lower()}"
            else:
                pattern = f"%{spec.value}"
            expr = col.like(pattern)
            return ~expr if spec.negated else expr

        return inner

    # String filter conditions for message
    @staticmethod
    def by_message_contains(spec: StringMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            col = cast(sa.ColumnElement[str | None], DeploymentHistoryRow.message)
            if spec.case_insensitive:
                col = sa.func.lower(col)
                pattern = f"%{spec.value.lower()}%"
            else:
                pattern = f"%{spec.value}%"
            expr = col.like(pattern)
            return ~expr if spec.negated else expr

        return inner

    @staticmethod
    def by_message_equals(spec: StringMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            col = cast(sa.ColumnElement[str | None], DeploymentHistoryRow.message)
            if spec.case_insensitive:
                col = sa.func.lower(col)
                val = spec.value.lower()
            else:
                val = spec.value
            if spec.negated:
                return col != val
            return col == val

        return inner

    @staticmethod
    def by_message_starts_with(spec: StringMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            col = cast(sa.ColumnElement[str | None], DeploymentHistoryRow.message)
            if spec.case_insensitive:
                col = sa.func.lower(col)
                pattern = f"{spec.value.lower()}%"
            else:
                pattern = f"{spec.value}%"
            expr = col.like(pattern)
            return ~expr if spec.negated else expr

        return inner

    @staticmethod
    def by_message_ends_with(spec: StringMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            col = cast(sa.ColumnElement[str | None], DeploymentHistoryRow.message)
            if spec.case_insensitive:
                col = sa.func.lower(col)
                pattern = f"%{spec.value.lower()}"
            else:
                pattern = f"%{spec.value}"
            expr = col.like(pattern)
            return ~expr if spec.negated else expr

        return inner

    @staticmethod
    def by_cursor_forward(cursor_id: str) -> QueryCondition:
        """Cursor condition for forward pagination (after cursor)."""
        cursor_uuid = uuid.UUID(cursor_id)

        def inner() -> sa.sql.expression.ColumnElement[bool]:
            subquery = (
                sa.select(DeploymentHistoryRow.created_at)
                .where(DeploymentHistoryRow.id == cursor_uuid)
                .scalar_subquery()
            )
            return DeploymentHistoryRow.created_at < subquery

        return inner

    @staticmethod
    def by_cursor_backward(cursor_id: str) -> QueryCondition:
        """Cursor condition for backward pagination (before cursor)."""
        cursor_uuid = uuid.UUID(cursor_id)

        def inner() -> sa.sql.expression.ColumnElement[bool]:
            subquery = (
                sa.select(DeploymentHistoryRow.created_at)
                .where(DeploymentHistoryRow.id == cursor_uuid)
                .scalar_subquery()
            )
            return DeploymentHistoryRow.created_at > subquery

        return inner

    # DateTime filter conditions
    @staticmethod
    def by_created_at_before(dt: datetime) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return DeploymentHistoryRow.created_at < dt

        return inner

    @staticmethod
    def by_created_at_after(dt: datetime) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return DeploymentHistoryRow.created_at > dt

        return inner

    @staticmethod
    def by_created_at_equals(dt: datetime) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return DeploymentHistoryRow.created_at == dt

        return inner

    @staticmethod
    def by_updated_at_before(dt: datetime) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return DeploymentHistoryRow.updated_at < dt

        return inner

    @staticmethod
    def by_updated_at_after(dt: datetime) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return DeploymentHistoryRow.updated_at > dt

        return inner

    @staticmethod
    def by_updated_at_equals(dt: datetime) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return DeploymentHistoryRow.updated_at == dt

        return inner


class DeploymentHistoryOrders:
    """Query orders for deployment history."""

    @staticmethod
    def created_at(ascending: bool = True) -> QueryOrder:
        if ascending:
            return DeploymentHistoryRow.created_at.asc()
        return DeploymentHistoryRow.created_at.desc()

    @staticmethod
    def updated_at(ascending: bool = True) -> QueryOrder:
        if ascending:
            return DeploymentHistoryRow.updated_at.asc()
        return DeploymentHistoryRow.updated_at.desc()


# ========== Route History ==========


class RouteHistoryConditions:
    """Query conditions for route history."""

    # UUID filter conditions for history id
    @staticmethod
    def by_id_filter(spec: UUIDEqualMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            if spec.negated:
                return RouteHistoryRow.id != spec.value
            return RouteHistoryRow.id == spec.value

        return inner

    @staticmethod
    def by_id_in(spec: UUIDInMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            if spec.negated:
                return RouteHistoryRow.id.notin_(spec.values)
            return RouteHistoryRow.id.in_(spec.values)

        return inner

    @staticmethod
    def by_ids(ids: Collection[uuid.UUID]) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return RouteHistoryRow.id.in_(ids)

        return inner

    @staticmethod
    def by_route_id(route_id: uuid.UUID) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return RouteHistoryRow.route_id == route_id

        return inner

    @staticmethod
    def by_deployment_id(deployment_id: uuid.UUID) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return RouteHistoryRow.deployment_id == deployment_id

        return inner

    @staticmethod
    def by_result(result: SchedulingResult) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return RouteHistoryRow.result == str(result)

        return inner

    @staticmethod
    def by_results(results: list[SchedulingResult]) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return RouteHistoryRow.result.in_([str(r) for r in results])

        return inner

    @staticmethod
    def by_from_status(status: RouteStatus) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return cast(
                sa.sql.expression.ColumnElement[bool], RouteHistoryRow.from_status == status.value
            )

        return inner

    @staticmethod
    def by_to_status(status: RouteStatus) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return cast(
                sa.sql.expression.ColumnElement[bool], RouteHistoryRow.to_status == status.value
            )

        return inner

    @staticmethod
    def by_error_code(error_code: str) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return RouteHistoryRow.error_code == error_code

        return inner

    # UUID filter conditions for route_id
    @staticmethod
    def by_route_id_filter(spec: UUIDEqualMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            if spec.negated:
                return RouteHistoryRow.route_id != spec.value
            return RouteHistoryRow.route_id == spec.value

        return inner

    @staticmethod
    def by_route_id_in(spec: UUIDInMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            if spec.negated:
                return RouteHistoryRow.route_id.notin_(spec.values)
            return RouteHistoryRow.route_id.in_(spec.values)

        return inner

    # UUID filter conditions for deployment_id
    @staticmethod
    def by_deployment_id_filter(spec: UUIDEqualMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            if spec.negated:
                return RouteHistoryRow.deployment_id != spec.value
            return RouteHistoryRow.deployment_id == spec.value

        return inner

    @staticmethod
    def by_deployment_id_in(spec: UUIDInMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            if spec.negated:
                return RouteHistoryRow.deployment_id.notin_(spec.values)
            return RouteHistoryRow.deployment_id.in_(spec.values)

        return inner

    # String filter conditions for error_code
    @staticmethod
    def by_error_code_contains(spec: StringMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            col = cast(sa.ColumnElement[str | None], RouteHistoryRow.error_code)
            if spec.case_insensitive:
                col = sa.func.lower(col)
                pattern = f"%{spec.value.lower()}%"
            else:
                pattern = f"%{spec.value}%"
            expr = col.like(pattern)
            return ~expr if spec.negated else expr

        return inner

    @staticmethod
    def by_error_code_equals(spec: StringMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            col = cast(sa.ColumnElement[str | None], RouteHistoryRow.error_code)
            if spec.case_insensitive:
                col = sa.func.lower(col)
                val = spec.value.lower()
            else:
                val = spec.value
            if spec.negated:
                return col != val
            return col == val

        return inner

    @staticmethod
    def by_error_code_starts_with(spec: StringMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            col = cast(sa.ColumnElement[str | None], RouteHistoryRow.error_code)
            if spec.case_insensitive:
                col = sa.func.lower(col)
                pattern = f"{spec.value.lower()}%"
            else:
                pattern = f"{spec.value}%"
            expr = col.like(pattern)
            return ~expr if spec.negated else expr

        return inner

    @staticmethod
    def by_error_code_ends_with(spec: StringMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            col = cast(sa.ColumnElement[str | None], RouteHistoryRow.error_code)
            if spec.case_insensitive:
                col = sa.func.lower(col)
                pattern = f"%{spec.value.lower()}"
            else:
                pattern = f"%{spec.value}"
            expr = col.like(pattern)
            return ~expr if spec.negated else expr

        return inner

    # Status list filter conditions
    @staticmethod
    def by_from_statuses(statuses: list[str]) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return RouteHistoryRow.from_status.in_(statuses)

        return inner

    @staticmethod
    def by_to_statuses(statuses: list[str]) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return RouteHistoryRow.to_status.in_(statuses)

        return inner

    # String filter conditions for phase
    @staticmethod
    def by_phase_contains(spec: StringMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            col = cast(sa.ColumnElement[str], RouteHistoryRow.phase)
            if spec.case_insensitive:
                col = sa.func.lower(col)
                pattern = f"%{spec.value.lower()}%"
            else:
                pattern = f"%{spec.value}%"
            expr = col.like(pattern)
            return ~expr if spec.negated else expr

        return inner

    @staticmethod
    def by_phase_equals(spec: StringMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            col = cast(sa.ColumnElement[str], RouteHistoryRow.phase)
            if spec.case_insensitive:
                col = sa.func.lower(col)
                val = spec.value.lower()
            else:
                val = spec.value
            if spec.negated:
                return col != val
            return col == val

        return inner

    @staticmethod
    def by_phase_starts_with(spec: StringMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            col = cast(sa.ColumnElement[str], RouteHistoryRow.phase)
            if spec.case_insensitive:
                col = sa.func.lower(col)
                pattern = f"{spec.value.lower()}%"
            else:
                pattern = f"{spec.value}%"
            expr = col.like(pattern)
            return ~expr if spec.negated else expr

        return inner

    @staticmethod
    def by_phase_ends_with(spec: StringMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            col = cast(sa.ColumnElement[str], RouteHistoryRow.phase)
            if spec.case_insensitive:
                col = sa.func.lower(col)
                pattern = f"%{spec.value.lower()}"
            else:
                pattern = f"%{spec.value}"
            expr = col.like(pattern)
            return ~expr if spec.negated else expr

        return inner

    # String filter conditions for message
    @staticmethod
    def by_message_contains(spec: StringMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            col = cast(sa.ColumnElement[str | None], RouteHistoryRow.message)
            if spec.case_insensitive:
                col = sa.func.lower(col)
                pattern = f"%{spec.value.lower()}%"
            else:
                pattern = f"%{spec.value}%"
            expr = col.like(pattern)
            return ~expr if spec.negated else expr

        return inner

    @staticmethod
    def by_message_equals(spec: StringMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            col = cast(sa.ColumnElement[str | None], RouteHistoryRow.message)
            if spec.case_insensitive:
                col = sa.func.lower(col)
                val = spec.value.lower()
            else:
                val = spec.value
            if spec.negated:
                return col != val
            return col == val

        return inner

    @staticmethod
    def by_message_starts_with(spec: StringMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            col = cast(sa.ColumnElement[str | None], RouteHistoryRow.message)
            if spec.case_insensitive:
                col = sa.func.lower(col)
                pattern = f"{spec.value.lower()}%"
            else:
                pattern = f"{spec.value}%"
            expr = col.like(pattern)
            return ~expr if spec.negated else expr

        return inner

    @staticmethod
    def by_message_ends_with(spec: StringMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            col = cast(sa.ColumnElement[str | None], RouteHistoryRow.message)
            if spec.case_insensitive:
                col = sa.func.lower(col)
                pattern = f"%{spec.value.lower()}"
            else:
                pattern = f"%{spec.value}"
            expr = col.like(pattern)
            return ~expr if spec.negated else expr

        return inner

    @staticmethod
    def by_cursor_forward(cursor_id: str) -> QueryCondition:
        """Cursor condition for forward pagination (after cursor)."""
        cursor_uuid = uuid.UUID(cursor_id)

        def inner() -> sa.sql.expression.ColumnElement[bool]:
            subquery = (
                sa.select(RouteHistoryRow.created_at)
                .where(RouteHistoryRow.id == cursor_uuid)
                .scalar_subquery()
            )
            return RouteHistoryRow.created_at < subquery

        return inner

    @staticmethod
    def by_cursor_backward(cursor_id: str) -> QueryCondition:
        """Cursor condition for backward pagination (before cursor)."""
        cursor_uuid = uuid.UUID(cursor_id)

        def inner() -> sa.sql.expression.ColumnElement[bool]:
            subquery = (
                sa.select(RouteHistoryRow.created_at)
                .where(RouteHistoryRow.id == cursor_uuid)
                .scalar_subquery()
            )
            return RouteHistoryRow.created_at > subquery

        return inner

    # DateTime filter conditions
    @staticmethod
    def by_created_at_before(dt: datetime) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return RouteHistoryRow.created_at < dt

        return inner

    @staticmethod
    def by_created_at_after(dt: datetime) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return RouteHistoryRow.created_at > dt

        return inner

    @staticmethod
    def by_created_at_equals(dt: datetime) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return RouteHistoryRow.created_at == dt

        return inner

    @staticmethod
    def by_updated_at_before(dt: datetime) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return RouteHistoryRow.updated_at < dt

        return inner

    @staticmethod
    def by_updated_at_after(dt: datetime) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return RouteHistoryRow.updated_at > dt

        return inner

    @staticmethod
    def by_updated_at_equals(dt: datetime) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return RouteHistoryRow.updated_at == dt

        return inner


class RouteHistoryOrders:
    """Query orders for route history."""

    @staticmethod
    def created_at(ascending: bool = True) -> QueryOrder:
        if ascending:
            return RouteHistoryRow.created_at.asc()
        return RouteHistoryRow.created_at.desc()

    @staticmethod
    def updated_at(ascending: bool = True) -> QueryOrder:
        if ascending:
            return RouteHistoryRow.updated_at.asc()
        return RouteHistoryRow.updated_at.desc()



from __future__ import annotations

from .actions import CreateErrorLogAction, CreateErrorLogActionResult
from .processors import ErrorLogProcessors
from .service import ErrorLogService

__all__ = (
    "CreateErrorLogAction",
    "CreateErrorLogActionResult",
    "ErrorLogProcessors",
    "ErrorLogService",
)



"""add preopen ports

Revision ID: 3bb80d1887d6
Revises: 1e8531583e20
Create Date: 2020-02-05 17:02:42.344726

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "3bb80d1887d6"
down_revision = "1e8531583e20"
branch_labels = None
depends_on = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column("kernels", sa.Column("preopen_ports", sa.ARRAY(sa.Integer()), nullable=True))
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("kernels", "preopen_ports")
    # ### end Alembic commands ###



from __future__ import annotations

from dataclasses import dataclass
from typing import override
from uuid import UUID

import sqlalchemy as sa

from ai.backend.manager.models.error_logs import ErrorLogRow
from ai.backend.manager.models.group import AssocGroupUserRow
from ai.backend.manager.models.keypair import KeyPairRow
from ai.backend.manager.models.user import UserRow
from ai.backend.manager.models.vfolder import VFolderPermissionRow
from ai.backend.manager.repositories.base.purger import BatchPurger, BatchPurgerSpec


@dataclass
class UserErrorLogBatchPurgerSpec(BatchPurgerSpec[ErrorLogRow]):
    """PurgerSpec for deleting all error logs belonging to a user."""

    user_uuid: UUID

    @override
    def build_subquery(self) -> sa.sql.Select[tuple[ErrorLogRow]]:
        return sa.select(ErrorLogRow).where(ErrorLogRow.__table__.c.user == self.user_uuid)


@dataclass
class UserKeyPairBatchPurgerSpec(BatchPurgerSpec[KeyPairRow]):
    """PurgerSpec for deleting all keypairs belonging to a user."""

    user_uuid: UUID

    @override
    def build_subquery(self) -> sa.sql.Select[tuple[KeyPairRow]]:
        return sa.select(KeyPairRow).where(KeyPairRow.user == self.user_uuid)


@dataclass
class UserVFolderPermissionBatchPurgerSpec(BatchPurgerSpec[VFolderPermissionRow]):
    """PurgerSpec for deleting all vfolder permissions belonging to a user."""

    user_uuid: UUID

    @override
    def build_subquery(self) -> sa.sql.Select[tuple[VFolderPermissionRow]]:
        return sa.select(VFolderPermissionRow).where(VFolderPermissionRow.user == self.user_uuid)


@dataclass
class UserGroupAssociationBatchPurgerSpec(BatchPurgerSpec[AssocGroupUserRow]):
    """PurgerSpec for deleting all group associations belonging to a user."""

    user_uuid: UUID

    @override
    def build_subquery(self) -> sa.sql.Select[tuple[AssocGroupUserRow]]:
        return sa.select(AssocGroupUserRow).where(AssocGroupUserRow.user_id == self.user_uuid)


@dataclass
class UserBatchPurgerSpec(BatchPurgerSpec[UserRow]):
    """PurgerSpec for deleting a user."""

    user_uuid: UUID

    @override
    def build_subquery(self) -> sa.sql.Select[tuple[UserRow]]:
        return sa.select(UserRow).where(UserRow.uuid == self.user_uuid)


def create_user_error_log_purger(user_uuid: UUID) -> BatchPurger[ErrorLogRow]:
    """Create a BatchPurger for deleting all error logs belonging to a user."""
    return BatchPurger(
        spec=UserErrorLogBatchPurgerSpec(user_uuid=user_uuid),
    )


def create_user_keypair_purger(user_uuid: UUID) -> BatchPurger[KeyPairRow]:
    """Create a BatchPurger for deleting all keypairs belonging to a user."""
    return BatchPurger(
        spec=UserKeyPairBatchPurgerSpec(user_uuid=user_uuid),
    )


def create_user_vfolder_permission_purger(user_uuid: UUID) -> BatchPurger[VFolderPermissionRow]:
    """Create a BatchPurger for deleting all vfolder permissions belonging to a user."""
    return BatchPurger(
        spec=UserVFolderPermissionBatchPurgerSpec(user_uuid=user_uuid),
    )


def create_user_group_association_purger(user_uuid: UUID) -> BatchPurger[AssocGroupUserRow]:
    """Create a BatchPurger for deleting all group associations belonging to a user."""
    return BatchPurger(
        spec=UserGroupAssociationBatchPurgerSpec(user_uuid=user_uuid),
    )


def create_user_purger(user_uuid: UUID) -> BatchPurger[UserRow]:
    """Create a BatchPurger for deleting a user."""
    return BatchPurger(
        spec=UserBatchPurgerSpec(user_uuid=user_uuid),
        batch_size=1,  # We expect only one row to be deleted
    )



"""Resource Usage Service."""

from __future__ import annotations

from ai.backend.manager.repositories.base import BatchQuerier
from ai.backend.manager.repositories.resource_usage_history import ResourceUsageHistoryRepository

from .actions import (
    SearchDomainUsageBucketsAction,
    SearchDomainUsageBucketsActionResult,
    SearchProjectUsageBucketsAction,
    SearchProjectUsageBucketsActionResult,
    SearchScopedDomainUsageBucketsAction,
    SearchScopedDomainUsageBucketsActionResult,
    SearchScopedProjectUsageBucketsAction,
    SearchScopedProjectUsageBucketsActionResult,
    SearchScopedUserUsageBucketsAction,
    SearchScopedUserUsageBucketsActionResult,
    SearchUserUsageBucketsAction,
    SearchUserUsageBucketsActionResult,
)

__all__ = ("ResourceUsageService",)


class ResourceUsageService:
    """Service for resource usage operations.

    Provides read operations wrapping the ResourceUsageHistoryRepository.
    Write operations (create/upsert) are handled directly by sokovan using the repository.
    """

    _repository: ResourceUsageHistoryRepository

    def __init__(self, repository: ResourceUsageHistoryRepository) -> None:
        self._repository = repository

    # Domain Usage Buckets

    async def search_domain_usage_buckets(
        self, action: SearchDomainUsageBucketsAction
    ) -> SearchDomainUsageBucketsActionResult:
        """Search domain usage buckets with pagination."""
        querier = BatchQuerier(
            pagination=action.pagination,
            conditions=action.conditions,
            orders=action.orders,
        )
        result = await self._repository.search_domain_usage_buckets(querier)
        return SearchDomainUsageBucketsActionResult(
            items=result.items,
            total_count=result.total_count,
        )

    # Project Usage Buckets

    async def search_project_usage_buckets(
        self, action: SearchProjectUsageBucketsAction
    ) -> SearchProjectUsageBucketsActionResult:
        """Search project usage buckets with pagination."""
        querier = BatchQuerier(
            pagination=action.pagination,
            conditions=action.conditions,
            orders=action.orders,
        )
        result = await self._repository.search_project_usage_buckets(querier)
        return SearchProjectUsageBucketsActionResult(
            items=result.items,
            total_count=result.total_count,
        )

    # User Usage Buckets

    async def search_user_usage_buckets(
        self, action: SearchUserUsageBucketsAction
    ) -> SearchUserUsageBucketsActionResult:
        """Search user usage buckets with pagination."""
        querier = BatchQuerier(
            pagination=action.pagination,
            conditions=action.conditions,
            orders=action.orders,
        )
        result = await self._repository.search_user_usage_buckets(querier)
        return SearchUserUsageBucketsActionResult(
            items=result.items,
            total_count=result.total_count,
        )

    # Scoped Usage Bucket Searches

    async def search_scoped_domain_usage_buckets(
        self,
        action: SearchScopedDomainUsageBucketsAction,
    ) -> SearchScopedDomainUsageBucketsActionResult:
        """Search domain usage buckets within scope."""
        result = await self._repository.search_domain_usage_buckets(
            action.querier,
            action.scope,
        )
        return SearchScopedDomainUsageBucketsActionResult(
            items=result.items,
            total_count=result.total_count,
        )

    async def search_scoped_project_usage_buckets(
        self,
        action: SearchScopedProjectUsageBucketsAction,
    ) -> SearchScopedProjectUsageBucketsActionResult:
        """Search project usage buckets within scope."""
        result = await self._repository.search_project_usage_buckets(
            action.querier,
            action.scope,
        )
        return SearchScopedProjectUsageBucketsActionResult(
            items=result.items,
            total_count=result.total_count,
        )

    async def search_scoped_user_usage_buckets(
        self,
        action: SearchScopedUserUsageBucketsAction,
    ) -> SearchScopedUserUsageBucketsActionResult:
        """Search user usage buckets within scope."""
        result = await self._repository.search_user_usage_buckets(
            action.querier,
            action.scope,
        )
        return SearchScopedUserUsageBucketsActionResult(
            items=result.items,
            total_count=result.total_count,
        )



"""Query conditions and orders for deployments."""

from __future__ import annotations

import uuid
from collections.abc import Collection
from typing import Any, cast

import sqlalchemy as sa

from ai.backend.common.data.endpoint.types import EndpointLifecycle
from ai.backend.common.data.model_deployment.types import ModelDeploymentStatus
from ai.backend.manager.api.gql.base import StringMatchSpec
from ai.backend.manager.models.endpoint import EndpointRow
from ai.backend.manager.repositories.base import QueryCondition, QueryOrder


class DeploymentConditions:
    """Query conditions for deployments."""

    @staticmethod
    def by_ids(deployment_ids: Collection[uuid.UUID]) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return EndpointRow.id.in_(deployment_ids)

        return inner

    @staticmethod
    def by_name_equals(spec: StringMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            if spec.case_insensitive:
                condition = sa.func.lower(EndpointRow.name) == spec.value.lower()
            else:
                condition = EndpointRow.name == spec.value
            if spec.negated:
                condition = sa.not_(condition)
            return condition

        return inner

    @staticmethod
    def by_name_contains(spec: StringMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            if spec.case_insensitive:
                condition = EndpointRow.name.ilike(f"%{spec.value}%")
            else:
                condition = EndpointRow.name.like(f"%{spec.value}%")
            if spec.negated:
                condition = sa.not_(condition)
            return condition

        return inner

    @staticmethod
    def by_name_starts_with(spec: StringMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            if spec.case_insensitive:
                condition = EndpointRow.name.ilike(f"{spec.value}%")
            else:
                condition = EndpointRow.name.like(f"{spec.value}%")
            if spec.negated:
                condition = sa.not_(condition)
            return condition

        return inner

    @staticmethod
    def by_name_ends_with(spec: StringMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            if spec.case_insensitive:
                condition = EndpointRow.name.ilike(f"%{spec.value}")
            else:
                condition = EndpointRow.name.like(f"%{spec.value}")
            if spec.negated:
                condition = sa.not_(condition)
            return condition

        return inner

    @staticmethod
    def by_project_id(project_id: uuid.UUID) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return EndpointRow.project == project_id

        return inner

    @staticmethod
    def by_domain_name_equals(spec: StringMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            if spec.case_insensitive:
                condition = sa.func.lower(EndpointRow.domain) == spec.value.lower()
            else:
                condition = EndpointRow.domain == spec.value
            if spec.negated:
                condition = sa.not_(condition)
            return condition

        return inner

    @staticmethod
    def by_domain_name_contains(spec: StringMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            if spec.case_insensitive:
                condition = EndpointRow.domain.ilike(f"%{spec.value}%")
            else:
                condition = EndpointRow.domain.like(f"%{spec.value}%")
            if spec.negated:
                condition = sa.not_(condition)
            return condition

        return inner

    @staticmethod
    def by_domain_name_starts_with(spec: StringMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            if spec.case_insensitive:
                condition = EndpointRow.domain.ilike(f"{spec.value}%")
            else:
                condition = EndpointRow.domain.like(f"{spec.value}%")
            if spec.negated:
                condition = sa.not_(condition)
            return condition

        return inner

    @staticmethod
    def by_domain_name_ends_with(spec: StringMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            if spec.case_insensitive:
                condition = EndpointRow.domain.ilike(f"%{spec.value}")
            else:
                condition = EndpointRow.domain.like(f"%{spec.value}")
            if spec.negated:
                condition = sa.not_(condition)
            return condition

        return inner

    @staticmethod
    def by_status_equals(status: ModelDeploymentStatus) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return EndpointRow.lifecycle_stage == status

        return inner

    @staticmethod
    def by_status_in(statuses: Collection[ModelDeploymentStatus]) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return EndpointRow.lifecycle_stage.in_(statuses)

        return inner

    @staticmethod
    def by_lifecycle_stages(statuses: Collection[EndpointLifecycle]) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return EndpointRow.lifecycle_stage.in_(statuses)

        return inner

    @staticmethod
    def by_open_to_public(value: bool) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return EndpointRow.open_to_public == value

        return inner

    @staticmethod
    def by_tag_contains(spec: StringMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            if spec.case_insensitive:
                condition = EndpointRow.tag.ilike(f"%{spec.value}%")
            else:
                condition = EndpointRow.tag.like(f"%{spec.value}%")
            if spec.negated:
                condition = sa.not_(condition)
            return condition

        return inner

    @staticmethod
    def by_tag_equals(spec: StringMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            if spec.case_insensitive:
                condition = sa.func.lower(EndpointRow.tag) == spec.value.lower()
            else:
                condition = EndpointRow.tag == spec.value
            if spec.negated:
                condition = sa.not_(condition)
            return condition

        return inner

    @staticmethod
    def by_tag_starts_with(spec: StringMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            if spec.case_insensitive:
                condition = EndpointRow.tag.ilike(f"{spec.value}%")
            else:
                condition = EndpointRow.tag.like(f"{spec.value}%")
            if spec.negated:
                condition = sa.not_(condition)
            return condition

        return inner

    @staticmethod
    def by_tag_ends_with(spec: StringMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            if spec.case_insensitive:
                condition = EndpointRow.tag.ilike(f"%{spec.value}")
            else:
                condition = EndpointRow.tag.like(f"%{spec.value}")
            if spec.negated:
                condition = sa.not_(condition)
            return condition

        return inner

    @staticmethod
    def by_url_contains(spec: StringMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            if spec.case_insensitive:
                condition = EndpointRow.url.ilike(f"%{spec.value}%")
            else:
                condition = EndpointRow.url.like(f"%{spec.value}%")
            if spec.negated:
                condition = sa.not_(condition)
            return condition

        return inner

    @staticmethod
    def by_url_equals(spec: StringMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            if spec.case_insensitive:
                condition = sa.func.lower(EndpointRow.url) == spec.value.lower()
            else:
                condition = EndpointRow.url == spec.value
            if spec.negated:
                condition = sa.not_(condition)
            return condition

        return inner

    @staticmethod
    def by_url_starts_with(spec: StringMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            if spec.case_insensitive:
                condition = EndpointRow.url.ilike(f"{spec.value}%")
            else:
                condition = EndpointRow.url.like(f"{spec.value}%")
            if spec.negated:
                condition = sa.not_(condition)
            return condition

        return inner

    @staticmethod
    def by_url_ends_with(spec: StringMatchSpec) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            if spec.case_insensitive:
                condition = EndpointRow.url.ilike(f"%{spec.value}")
            else:
                condition = EndpointRow.url.like(f"%{spec.value}")
            if spec.negated:
                condition = sa.not_(condition)
            return condition

        return inner

    @staticmethod
    def by_cursor_forward(cursor_id: str) -> QueryCondition:
        """Cursor condition for forward pagination (after cursor).

        Uses subquery to get created_at of the cursor row and compare.
        """
        cursor_uuid = uuid.UUID(cursor_id)

        def inner() -> sa.sql.expression.ColumnElement[bool]:
            subquery = (
                sa.select(EndpointRow.created_at)
                .where(EndpointRow.id == cursor_uuid)
                .scalar_subquery()
            )
            return EndpointRow.created_at < subquery

        return inner

    @staticmethod
    def by_cursor_backward(cursor_id: str) -> QueryCondition:
        """Cursor condition for backward pagination (before cursor).

        Uses subquery to get created_at of the cursor row and compare.
        """
        cursor_uuid = uuid.UUID(cursor_id)

        def inner() -> sa.sql.expression.ColumnElement[bool]:
            subquery = (
                sa.select(EndpointRow.created_at)
                .where(EndpointRow.id == cursor_uuid)
                .scalar_subquery()
            )
            return EndpointRow.created_at > subquery

        return inner


class DeploymentOrders:
    """Query orders for deployments."""

    @staticmethod
    def name(ascending: bool = True) -> QueryOrder:
        if ascending:
            return EndpointRow.name.asc()
        return EndpointRow.name.desc()

    @staticmethod
    def created_at(ascending: bool = True) -> QueryOrder:
        if ascending:
            return EndpointRow.created_at.asc()
        return EndpointRow.created_at.desc()

    @staticmethod
    def updated_at(ascending: bool = True) -> QueryOrder:
        if ascending:
            return cast(
                sa.UnaryExpression[Any] | sa.ColumnElement[Any], EndpointRow.updated_at.asc()
            )
        return cast(sa.UnaryExpression[Any] | sa.ColumnElement[Any], EndpointRow.updated_at.desc())



"""rename_scaling_group_to_resource_group

Revision ID: a1b2c3d4e5f6
Revises: f8a9b3c2d1e0
Create Date: 2026-01-15 10:00:00.000000

"""

from alembic import op

# revision identifiers, used by Alembic.
revision = "a1b2c3d4e5f6"
down_revision = "f8a9b3c2d1e0"
branch_labels = None
depends_on = None


def upgrade() -> None:
    # Rename columns in all fair_share and resource_usage tables
    # scaling_group → resource_group

    # === domain_fair_shares ===
    # Drop indexes that reference scaling_group
    op.drop_index("ix_domain_fair_shares_scaling_group", table_name="domain_fair_shares")
    op.drop_index("ix_domain_fair_share_lookup", table_name="domain_fair_shares")
    # Drop unique constraint
    op.drop_constraint("uq_domain_fair_share", table_name="domain_fair_shares", type_="unique")
    # Rename column
    op.alter_column("domain_fair_shares", "scaling_group", new_column_name="resource_group")
    # Recreate indexes with new column name
    op.create_index(
        op.f("ix_domain_fair_shares_resource_group"),
        "domain_fair_shares",
        ["resource_group"],
        unique=False,
    )
    op.create_index(
        "ix_domain_fair_share_lookup",
        "domain_fair_shares",
        ["resource_group", "domain_name"],
        unique=False,
    )
    # Recreate unique constraint
    op.create_unique_constraint(
        "uq_domain_fair_share", "domain_fair_shares", ["resource_group", "domain_name"]
    )

    # === project_fair_shares ===
    op.drop_index("ix_project_fair_shares_scaling_group", table_name="project_fair_shares")
    op.drop_index("ix_project_fair_share_lookup", table_name="project_fair_shares")
    op.drop_constraint("uq_project_fair_share", table_name="project_fair_shares", type_="unique")
    op.alter_column("project_fair_shares", "scaling_group", new_column_name="resource_group")
    op.create_index(
        op.f("ix_project_fair_shares_resource_group"),
        "project_fair_shares",
        ["resource_group"],
        unique=False,
    )
    op.create_index(
        "ix_project_fair_share_lookup",
        "project_fair_shares",
        ["resource_group", "project_id"],
        unique=False,
    )
    op.create_unique_constraint(
        "uq_project_fair_share", "project_fair_shares", ["resource_group", "project_id"]
    )

    # === user_fair_shares ===
    op.drop_index("ix_user_fair_shares_scaling_group", table_name="user_fair_shares")
    op.drop_index("ix_user_fair_share_lookup", table_name="user_fair_shares")
    op.drop_constraint("uq_user_fair_share", table_name="user_fair_shares", type_="unique")
    op.alter_column("user_fair_shares", "scaling_group", new_column_name="resource_group")
    op.create_index(
        op.f("ix_user_fair_shares_resource_group"),
        "user_fair_shares",
        ["resource_group"],
        unique=False,
    )
    op.create_index(
        "ix_user_fair_share_lookup",
        "user_fair_shares",
        ["resource_group", "user_uuid", "project_id"],
        unique=False,
    )
    op.create_unique_constraint(
        "uq_user_fair_share", "user_fair_shares", ["resource_group", "user_uuid", "project_id"]
    )

    # === domain_usage_buckets ===
    op.drop_index("ix_domain_usage_bucket_lookup", table_name="domain_usage_buckets")
    op.drop_constraint("uq_domain_usage_bucket", table_name="domain_usage_buckets", type_="unique")
    op.alter_column("domain_usage_buckets", "scaling_group", new_column_name="resource_group")
    op.create_index(
        "ix_domain_usage_bucket_lookup",
        "domain_usage_buckets",
        ["domain_name", "resource_group", "period_start"],
        unique=False,
    )
    op.create_unique_constraint(
        "uq_domain_usage_bucket",
        "domain_usage_buckets",
        ["domain_name", "resource_group", "period_start"],
    )

    # === project_usage_buckets ===
    op.drop_index("ix_project_usage_bucket_lookup", table_name="project_usage_buckets")
    op.drop_constraint(
        "uq_project_usage_bucket", table_name="project_usage_buckets", type_="unique"
    )
    op.alter_column("project_usage_buckets", "scaling_group", new_column_name="resource_group")
    op.create_index(
        "ix_project_usage_bucket_lookup",
        "project_usage_buckets",
        ["project_id", "resource_group", "period_start"],
        unique=False,
    )
    op.create_unique_constraint(
        "uq_project_usage_bucket",
        "project_usage_buckets",
        ["project_id", "resource_group", "period_start"],
    )

    # === user_usage_buckets ===
    op.drop_index("ix_user_usage_bucket_lookup", table_name="user_usage_buckets")
    op.drop_constraint("uq_user_usage_bucket", table_name="user_usage_buckets", type_="unique")
    op.alter_column("user_usage_buckets", "scaling_group", new_column_name="resource_group")
    op.create_index(
        "ix_user_usage_bucket_lookup",
        "user_usage_buckets",
        ["user_uuid", "project_id", "resource_group", "period_start"],
        unique=False,
    )
    op.create_unique_constraint(
        "uq_user_usage_bucket",
        "user_usage_buckets",
        ["user_uuid", "project_id", "resource_group", "period_start"],
    )

    # === kernel_usage_records ===
    op.drop_index("ix_kernel_usage_records_scaling_group", table_name="kernel_usage_records")
    op.drop_index("ix_kernel_usage_sg_period", table_name="kernel_usage_records")
    op.alter_column("kernel_usage_records", "scaling_group", new_column_name="resource_group")
    op.create_index(
        op.f("ix_kernel_usage_records_resource_group"),
        "kernel_usage_records",
        ["resource_group"],
        unique=False,
    )
    op.create_index(
        "ix_kernel_usage_rg_period",
        "kernel_usage_records",
        ["resource_group", "period_start"],
        unique=False,
    )


def downgrade() -> None:
    # Reverse all changes: resource_group → scaling_group

    # === kernel_usage_records ===
    op.drop_index("ix_kernel_usage_rg_period", table_name="kernel_usage_records")
    op.drop_index("ix_kernel_usage_records_resource_group", table_name="kernel_usage_records")
    op.alter_column("kernel_usage_records", "resource_group", new_column_name="scaling_group")
    op.create_index(
        "ix_kernel_usage_sg_period",
        "kernel_usage_records",
        ["scaling_group", "period_start"],
        unique=False,
    )
    op.create_index(
        op.f("ix_kernel_usage_records_scaling_group"),
        "kernel_usage_records",
        ["scaling_group"],
        unique=False,
    )

    # === user_usage_buckets ===
    op.drop_constraint("uq_user_usage_bucket", table_name="user_usage_buckets", type_="unique")
    op.drop_index("ix_user_usage_bucket_lookup", table_name="user_usage_buckets")
    op.alter_column("user_usage_buckets", "resource_group", new_column_name="scaling_group")
    op.create_index(
        "ix_user_usage_bucket_lookup",
        "user_usage_buckets",
        ["user_uuid", "project_id", "scaling_group", "period_start"],
        unique=False,
    )
    op.create_unique_constraint(
        "uq_user_usage_bucket",
        "user_usage_buckets",
        ["user_uuid", "project_id", "scaling_group", "period_start"],
    )

    # === project_usage_buckets ===
    op.drop_constraint(
        "uq_project_usage_bucket", table_name="project_usage_buckets", type_="unique"
    )
    op.drop_index("ix_project_usage_bucket_lookup", table_name="project_usage_buckets")
    op.alter_column("project_usage_buckets", "resource_group", new_column_name="scaling_group")
    op.create_index(
        "ix_project_usage_bucket_lookup",
        "project_usage_buckets",
        ["project_id", "scaling_group", "period_start"],
        unique=False,
    )
    op.create_unique_constraint(
        "uq_project_usage_bucket",
        "project_usage_buckets",
        ["project_id", "scaling_group", "period_start"],
    )

    # === domain_usage_buckets ===
    op.drop_constraint("uq_domain_usage_bucket", table_name="domain_usage_buckets", type_="unique")
    op.drop_index("ix_domain_usage_bucket_lookup", table_name="domain_usage_buckets")
    op.alter_column("domain_usage_buckets", "resource_group", new_column_name="scaling_group")
    op.create_index(
        "ix_domain_usage_bucket_lookup",
        "domain_usage_buckets",
        ["domain_name", "scaling_group", "period_start"],
        unique=False,
    )
    op.create_unique_constraint(
        "uq_domain_usage_bucket",
        "domain_usage_buckets",
        ["domain_name", "scaling_group", "period_start"],
    )

    # === user_fair_shares ===
    op.drop_constraint("uq_user_fair_share", table_name="user_fair_shares", type_="unique")
    op.drop_index("ix_user_fair_share_lookup", table_name="user_fair_shares")
    op.drop_index("ix_user_fair_shares_resource_group", table_name="user_fair_shares")
    op.alter_column("user_fair_shares", "resource_group", new_column_name="scaling_group")
    op.create_index(
        "ix_user_fair_share_lookup",
        "user_fair_shares",
        ["scaling_group", "user_uuid", "project_id"],
        unique=False,
    )
    op.create_index(
        op.f("ix_user_fair_shares_scaling_group"),
        "user_fair_shares",
        ["scaling_group"],
        unique=False,
    )
    op.create_unique_constraint(
        "uq_user_fair_share", "user_fair_shares", ["scaling_group", "user_uuid", "project_id"]
    )

    # === project_fair_shares ===
    op.drop_constraint("uq_project_fair_share", table_name="project_fair_shares", type_="unique")
    op.drop_index("ix_project_fair_share_lookup", table_name="project_fair_shares")
    op.drop_index("ix_project_fair_shares_resource_group", table_name="project_fair_shares")
    op.alter_column("project_fair_shares", "resource_group", new_column_name="scaling_group")
    op.create_index(
        "ix_project_fair_share_lookup",
        "project_fair_shares",
        ["scaling_group", "project_id"],
        unique=False,
    )
    op.create_index(
        op.f("ix_project_fair_shares_scaling_group"),
        "project_fair_shares",
        ["scaling_group"],
        unique=False,
    )
    op.create_unique_constraint(
        "uq_project_fair_share", "project_fair_shares", ["scaling_group", "project_id"]
    )

    # === domain_fair_shares ===
    op.drop_constraint("uq_domain_fair_share", table_name="domain_fair_shares", type_="unique")
    op.drop_index("ix_domain_fair_share_lookup", table_name="domain_fair_shares")
    op.drop_index("ix_domain_fair_shares_resource_group", table_name="domain_fair_shares")
    op.alter_column("domain_fair_shares", "resource_group", new_column_name="scaling_group")
    op.create_index(
        "ix_domain_fair_share_lookup",
        "domain_fair_shares",
        ["scaling_group", "domain_name"],
        unique=False,
    )
    op.create_index(
        op.f("ix_domain_fair_shares_scaling_group"),
        "domain_fair_shares",
        ["scaling_group"],
        unique=False,
    )
    op.create_unique_constraint(
        "uq_domain_fair_share", "domain_fair_shares", ["scaling_group", "domain_name"]
    )



import enum
from abc import ABC, abstractmethod
from dataclasses import dataclass

from ai.backend.common.data.permission.types import EntityType, OperationType


class OperationStatus(enum.StrEnum):
    SUCCESS = "success"
    ERROR = "error"
    UNKNOWN = "unknown"
    RUNNING = "running"


class ActionOperationType(enum.StrEnum):
    GET = "get"
    SEARCH = "search"
    CREATE = "create"
    UPDATE = "update"
    DELETE = "delete"
    PURGE = "purge"

    def to_permission_operation(self) -> OperationType:
        match self:
            case ActionOperationType.GET:
                return OperationType.READ
            case ActionOperationType.SEARCH:
                return OperationType.READ
            case ActionOperationType.CREATE:
                return OperationType.CREATE
            case ActionOperationType.UPDATE:
                return OperationType.UPDATE
            case ActionOperationType.DELETE:
                return OperationType.SOFT_DELETE
            case ActionOperationType.PURGE:
                return OperationType.HARD_DELETE


@dataclass
class ActionSpec:
    entity_type: EntityType
    operation_type: ActionOperationType

    def type(self) -> str:
        return f"{self.entity_type}:{self.operation_type}"


class AbstractProcessorPackage(ABC):
    @abstractmethod
    def supported_actions(self) -> list[ActionSpec]:
        """Get the list of action specs that this processors can handle."""
        raise NotImplementedError



from typing import TYPE_CHECKING

from ai.backend.manager.container_registry import get_container_registry_cls
from ai.backend.manager.data.container_registry.types import ContainerRegistryData
from ai.backend.manager.errors.image import ContainerRegistryNotFound
from ai.backend.manager.models.utils import ExtendedAsyncSAEngine
from ai.backend.manager.repositories.container_registry.repository import (
    ContainerRegistryRepository,
)
from ai.backend.manager.services.container_registry.actions.clear_images import (
    ClearImagesAction,
    ClearImagesActionResult,
)
from ai.backend.manager.services.container_registry.actions.create_container_registry import (
    CreateContainerRegistryAction,
    CreateContainerRegistryActionResult,
)
from ai.backend.manager.services.container_registry.actions.delete_container_registry import (
    DeleteContainerRegistryAction,
    DeleteContainerRegistryActionResult,
)
from ai.backend.manager.services.container_registry.actions.get_container_registries import (
    GetContainerRegistriesAction,
    GetContainerRegistriesActionResult,
)
from ai.backend.manager.services.container_registry.actions.load_all_container_registries import (
    LoadAllContainerRegistriesAction,
    LoadAllContainerRegistriesActionResult,
)
from ai.backend.manager.services.container_registry.actions.load_container_registries import (
    LoadContainerRegistriesAction,
    LoadContainerRegistriesActionResult,
)
from ai.backend.manager.services.container_registry.actions.modify_container_registry import (
    ModifyContainerRegistryAction,
    ModifyContainerRegistryActionResult,
)
from ai.backend.manager.services.container_registry.actions.rescan_images import (
    RescanImagesAction,
    RescanImagesActionResult,
)

if TYPE_CHECKING:
    from ai.backend.manager.models.container_registry import ContainerRegistryRow


class ContainerRegistryService:
    _db: ExtendedAsyncSAEngine
    _container_registry_repository: ContainerRegistryRepository

    def __init__(
        self,
        db: ExtendedAsyncSAEngine,
        container_registry_repository: ContainerRegistryRepository,
    ) -> None:
        self._db = db
        self._container_registry_repository = container_registry_repository

    async def create_container_registry(
        self, action: CreateContainerRegistryAction
    ) -> CreateContainerRegistryActionResult:
        data = await self._container_registry_repository.create_registry(action.creator)
        return CreateContainerRegistryActionResult(data=data)

    async def modify_container_registry(
        self, action: ModifyContainerRegistryAction
    ) -> ModifyContainerRegistryActionResult:
        data = await self._container_registry_repository.modify_registry(action.updater)
        return ModifyContainerRegistryActionResult(data=data)

    async def delete_container_registry(
        self, action: DeleteContainerRegistryAction
    ) -> DeleteContainerRegistryActionResult:
        data = await self._container_registry_repository.delete_registry(action.purger)
        return DeleteContainerRegistryActionResult(data=data)

    async def rescan_images(self, action: RescanImagesAction) -> RescanImagesActionResult:
        registry_name = action.registry
        project = action.project

        registry_row: ContainerRegistryRow = (
            await self._container_registry_repository.get_registry_row_for_scanner(
                registry_name, project
            )
        )

        scanner_cls = get_container_registry_cls(registry_row)
        scanner = scanner_cls(self._db, registry_name, registry_row)
        result = await scanner.rescan_single_registry(action.progress_reporter)

        return RescanImagesActionResult(
            images=result.images, errors=result.errors, registry=registry_row.to_dataclass()
        )

    async def clear_images(self, action: ClearImagesAction) -> ClearImagesActionResult:
        registry_data = await self._container_registry_repository.clear_images(
            action.registry, action.project
        )

        return ClearImagesActionResult(registry=registry_data)

    async def load_container_registries(
        self, action: LoadContainerRegistriesAction
    ) -> LoadContainerRegistriesActionResult:
        registries: list[ContainerRegistryData] = []
        if action.project is not None:
            try:
                registry_data = (
                    await self._container_registry_repository.get_by_registry_and_project(
                        action.registry, action.project
                    )
                )
                registries = [registry_data]
            except ContainerRegistryNotFound:
                registries = []
        else:
            registries = await self._container_registry_repository.get_by_registry_name(
                action.registry
            )

        return LoadContainerRegistriesActionResult(registries=registries)

    async def load_all_container_registries(
        self, _action: LoadAllContainerRegistriesAction
    ) -> LoadAllContainerRegistriesActionResult:
        registries = await self._container_registry_repository.get_all()
        return LoadAllContainerRegistriesActionResult(registries=registries)

    async def get_container_registries(
        self, _action: GetContainerRegistriesAction
    ) -> GetContainerRegistriesActionResult:
        registries = await self._container_registry_repository.get_known_registries()
        return GetContainerRegistriesActionResult(registries=registries)



"""ProjectV2 GraphQL types package."""

from .enums import ProjectTypeEnum, VFolderHostPermissionEnum
from .filters import (
    ProjectDomainNestedFilter,
    ProjectUserNestedFilter,
    ProjectV2Filter,
    ProjectV2OrderBy,
    ProjectV2OrderField,
)
from .nested import (
    ProjectBasicInfoGQL,
    ProjectLifecycleInfoGQL,
    ProjectOrganizationInfoGQL,
    ProjectStorageInfoGQL,
    VFolderHostPermissionEntryGQL,
)
from .node import ProjectV2Connection, ProjectV2Edge, ProjectV2GQL
from .scopes import DomainProjectScope

__all__ = [
    # Enums
    "ProjectTypeEnum",
    "VFolderHostPermissionEnum",
    # Filters and OrderBy
    "ProjectDomainNestedFilter",
    "ProjectUserNestedFilter",
    "ProjectV2Filter",
    "ProjectV2OrderBy",
    "ProjectV2OrderField",
    # Scopes
    "DomainProjectScope",
    # Nested types - Basic
    "ProjectBasicInfoGQL",
    # Nested types - Organization
    "ProjectOrganizationInfoGQL",
    # Nested types - Storage
    "VFolderHostPermissionEntryGQL",
    "ProjectStorageInfoGQL",
    # Nested types - Lifecycle
    "ProjectLifecycleInfoGQL",
    # Node types
    "ProjectV2GQL",
    "ProjectV2Edge",
    "ProjectV2Connection",
]



"""Image-related data types."""

from __future__ import annotations

from dataclasses import dataclass

from ai.backend.common.types import AutoPullBehavior, ImageConfig


@dataclass(frozen=True)
class ImageIdentifier:
    """Identifier for an image with architecture."""

    image: str
    architecture: str


@dataclass
class ImageConfigData:
    """Image configuration data resolved from database."""

    canonical: str
    architecture: str
    project: str | None
    is_local: bool
    digest: str
    labels: dict[str, str]
    registry_name: str
    registry_url: str
    registry_username: str | None
    registry_password: str | None

    def to_image_config(self, auto_pull: AutoPullBehavior) -> ImageConfig:
        """
        Convert ImageConfigData to ImageConfig format for agents.

        :param auto_pull: Auto pull behavior setting
        :return: ImageConfig dictionary for agent RPC calls
        """
        return ImageConfig(
            architecture=self.architecture,
            project=self.project,
            canonical=self.canonical,
            is_local=self.is_local,
            digest=self.digest,
            labels=self.labels,
            repo_digest=None,
            registry={
                "name": self.registry_name,
                "url": self.registry_url,
                "username": self.registry_username,
                "password": self.registry_password,
            },
            auto_pull=auto_pull,
        )



"""
Provisioner plugins for sokovan scheduler.

This package contains the core plugin components for session scheduling:
- validators: Resource constraint validation rules
- sequencers: Session ordering/priority algorithms
- selectors: Agent selection strategies
- allocators: Resource allocation implementations
- provisioner: Main SessionProvisioner orchestrating the PENDING -> SCHEDULED transition
"""

from .provisioner import SessionProvisioner, SessionProvisionerArgs

__all__ = [
    "SessionProvisioner",
    "SessionProvisionerArgs",
]



from __future__ import annotations

import uuid
from collections.abc import Sequence

from ai.backend.manager.data.notification import NotificationChannelData, NotificationRuleData
from ai.backend.manager.repositories.base import BatchQuerier, OffsetPagination
from ai.backend.manager.repositories.notification.options import (
    NotificationChannelConditions,
    NotificationRuleConditions,
)
from ai.backend.manager.services.notification.actions.list_channels import SearchChannelsAction
from ai.backend.manager.services.notification.actions.list_rules import SearchRulesAction
from ai.backend.manager.services.notification.processors import NotificationProcessors


async def load_channels_by_ids(
    processor: NotificationProcessors,
    channel_ids: Sequence[uuid.UUID],
) -> list[NotificationChannelData | None]:
    """Batch load notification channels by their IDs.

    Args:
        processor: The notification processor.
        channel_ids: Sequence of channel UUIDs to load.

    Returns:
        List of NotificationChannelData (or None if not found) in the same order as channel_ids.
    """
    if not channel_ids:
        return []

    querier = BatchQuerier(
        pagination=OffsetPagination(limit=len(channel_ids)),
        conditions=[NotificationChannelConditions.by_ids(channel_ids)],
    )

    action_result = await processor.search_channels.wait_for_complete(
        SearchChannelsAction(querier=querier)
    )

    channel_map = {channel.id: channel for channel in action_result.channels}
    return [channel_map.get(channel_id) for channel_id in channel_ids]


async def load_rules_by_ids(
    processor: NotificationProcessors,
    rule_ids: Sequence[uuid.UUID],
) -> list[NotificationRuleData | None]:
    """Batch load notification rules by their IDs.

    Args:
        processor: The notification processor.
        rule_ids: Sequence of rule UUIDs to load.

    Returns:
        List of NotificationRuleData (or None if not found) in the same order as rule_ids.
    """
    if not rule_ids:
        return []

    querier = BatchQuerier(
        pagination=OffsetPagination(limit=len(rule_ids)),
        conditions=[NotificationRuleConditions.by_ids(rule_ids)],
    )

    action_result = await processor.search_rules.wait_for_complete(
        SearchRulesAction(querier=querier)
    )

    rule_map = {rule.id: rule for rule in action_result.rules}
    return [rule_map.get(rule_id) for rule_id in rule_ids]



"""CreatorSpec implementations for notification repository."""

from __future__ import annotations

from dataclasses import dataclass
from typing import override
from uuid import UUID

from ai.backend.common.data.notification import (
    NotificationChannelType,
    NotificationRuleType,
    WebhookSpec,
)
from ai.backend.common.data.notification.types import EmailSpec
from ai.backend.manager.models.notification import NotificationChannelRow, NotificationRuleRow
from ai.backend.manager.repositories.base import CreatorSpec


@dataclass
class NotificationChannelCreatorSpec(CreatorSpec[NotificationChannelRow]):
    """CreatorSpec for notification channel."""

    name: str
    channel_type: NotificationChannelType
    spec: WebhookSpec | EmailSpec
    created_by: UUID
    description: str | None = None
    enabled: bool = True

    @override
    def build_row(self) -> NotificationChannelRow:
        return NotificationChannelRow(
            name=self.name,
            description=self.description,
            channel_type=str(self.channel_type),
            config=self.spec.model_dump(),
            enabled=self.enabled,
            created_by=self.created_by,
        )


@dataclass
class NotificationRuleCreatorSpec(CreatorSpec[NotificationRuleRow]):
    """CreatorSpec for notification rule."""

    name: str
    rule_type: NotificationRuleType
    channel_id: UUID
    message_template: str
    created_by: UUID
    description: str | None = None
    enabled: bool = True

    @override
    def build_row(self) -> NotificationRuleRow:
        return NotificationRuleRow(
            name=self.name,
            description=self.description,
            rule_type=str(self.rule_type),
            channel_id=self.channel_id,
            message_template=self.message_template,
            enabled=self.enabled,
            created_by=self.created_by,
        )



from __future__ import annotations

from collections.abc import Mapping
from dataclasses import dataclass
from datetime import datetime
from enum import StrEnum
from functools import lru_cache
from typing import TYPE_CHECKING, Any
from uuid import UUID

from ai.backend.common.types import (
    CIStrEnum,
    KernelId,
    ResourceSlot,
    SessionId,
    SessionResult,
    SessionTypes,
)
from ai.backend.manager.data.image.types import ImageIdentifier

if TYPE_CHECKING:
    from ai.backend.manager.data.session.types import SchedulingResult


class KernelStatus(CIStrEnum):
    # values are only meaningful inside the manager
    PENDING = "PENDING"
    # ---
    SCHEDULED = "SCHEDULED"
    PREPARING = "PREPARING"
    # ---
    BUILDING = "BUILDING"
    PULLING = "PULLING"
    PREPARED = "PREPARED"
    CREATING = "CREATING"
    # ---
    RUNNING = "RUNNING"
    RESTARTING = "RESTARTING"
    RESIZING = "RESIZING"
    SUSPENDED = "SUSPENDED"
    # ---
    TERMINATING = "TERMINATING"
    TERMINATED = "TERMINATED"
    ERROR = "ERROR"
    CANCELLED = "CANCELLED"

    @classmethod
    def having_containers(cls) -> set[KernelStatus]:
        return {
            cls.PULLING,
            cls.CREATING,
            cls.RUNNING,
        }

    def have_container(self) -> bool:
        """
        Check if the current status is one of the statuses that have containers.
        """
        return self in KernelStatus.having_containers()

    @classmethod
    @lru_cache(maxsize=1)
    def resource_occupied_statuses(cls) -> frozenset[KernelStatus]:
        """
        Returns a set of kernel statuses that are considered as resource-occupying.
        """
        return frozenset((
            cls.RUNNING,
            cls.TERMINATING,
        ))

    @classmethod
    @lru_cache(maxsize=1)
    def resource_requested_statuses(cls) -> frozenset[KernelStatus]:
        """
        Returns a set of kernel statuses that are considered as resource-occupying.
        """
        return frozenset((
            cls.SCHEDULED,
            cls.PREPARING,
            cls.PULLING,
            cls.PREPARED,
            cls.CREATING,
        ))

    @classmethod
    @lru_cache(maxsize=1)
    def terminatable_statuses(cls) -> frozenset[KernelStatus]:
        """Return statuses that can transition to TERMINATING."""
        return frozenset(
            status
            for status in cls
            if status
            not in (
                cls.PENDING,
                cls.TERMINATING,
                cls.TERMINATED,
                cls.CANCELLED,
                cls.ERROR,
            )
        )

    @classmethod
    @lru_cache(maxsize=1)
    def terminal_statuses(cls) -> frozenset[KernelStatus]:
        """
        Returns a set of kernel statuses that are considered terminal.
        """
        return frozenset((
            cls.ERROR,
            cls.TERMINATED,
            cls.CANCELLED,
        ))

    @classmethod
    @lru_cache(maxsize=1)
    def pre_prepared_statuses(cls) -> frozenset[KernelStatus]:
        """
        Returns statuses before image pulling is complete.
        Used for NOT_ANY kernel matching to promote sessions to PREPARED.
        """
        return frozenset((
            cls.PENDING,
            cls.SCHEDULED,
            cls.PREPARING,
            cls.BUILDING,
            cls.PULLING,
        ))

    @classmethod
    @lru_cache(maxsize=1)
    def pre_running_statuses(cls) -> frozenset[KernelStatus]:
        """
        Returns statuses before kernel is actually running.
        Used for NOT_ANY kernel matching to promote sessions to RUNNING.
        """
        return frozenset((
            cls.PENDING,
            cls.SCHEDULED,
            cls.PREPARING,
            cls.BUILDING,
            cls.PULLING,
            cls.PREPARED,
            cls.CREATING,
        ))

    @classmethod
    @lru_cache(maxsize=1)
    def retriable_statuses(cls) -> frozenset[KernelStatus]:
        """
        Returns a set of kernel statuses that are considered retriable.
        """
        return frozenset(
            status
            for status in cls
            if status
            not in (
                cls.RUNNING,
                cls.TERMINATING,
                cls.TERMINATED,
                cls.CANCELLED,
            )
        )


@dataclass
class RelatedSessionInfo:
    session_id: str  # Session UUID
    creation_id: str | None
    name: str | None
    session_type: SessionTypes


@dataclass
class ClusterConfig:
    cluster_mode: str
    cluster_size: int
    cluster_role: str
    cluster_idx: int
    local_rank: int
    cluster_hostname: str


@dataclass
class UserPermission:
    user_uuid: UUID
    access_key: str
    domain_name: str
    group_id: UUID
    uid: int | None
    main_gid: int | None
    gids: list[int] | None


@dataclass
class ResourceInfo:
    scaling_group: str | None
    agent: str | None
    agent_addr: str | None
    container_id: str | None
    occupied_slots: ResourceSlot
    requested_slots: ResourceSlot
    occupied_shares: dict[str, Any]
    attached_devices: dict[str, Any]
    resource_opts: dict[str, Any]


@dataclass
class ImageInfo:
    identifier: ImageIdentifier | None
    registry: str | None
    tag: str | None
    architecture: str | None


@dataclass
class RuntimeConfig:
    environ: list[str] | None
    mounts: list[str] | None  # legacy
    mount_map: dict[str, Any] | None  # legacy
    vfolder_mounts: list[dict[str, Any]] | None
    bootstrap_script: str | None
    startup_command: str | None


@dataclass
class NetworkConfig:
    kernel_host: str | None
    repl_in_port: int
    repl_out_port: int
    stdin_port: int  # legacy
    stdout_port: int  # legacy
    service_ports: list[dict[str, Any]] | None
    preopen_ports: list[int] | None
    use_host_network: bool


@dataclass
class LifecycleStatus:
    status: KernelStatus
    result: SessionResult
    created_at: datetime | None
    terminated_at: datetime | None
    starts_at: datetime | None
    status_changed: datetime | None
    status_info: str | None
    status_data: Mapping[str, Any] | None
    status_history: dict[str, Any] | None
    last_seen: datetime | None
    last_observed_at: datetime | None


@dataclass
class Metrics:
    num_queries: int
    last_stat: dict[str, Any] | None
    container_log: bytes | None


@dataclass
class Metadata:
    callback_url: str | None
    internal_data: dict[str, Any] | None


@dataclass
class KernelInfo:
    id: KernelId
    session: RelatedSessionInfo
    user_permission: UserPermission
    image: ImageInfo
    network: NetworkConfig
    cluster: ClusterConfig
    resource: ResourceInfo
    runtime: RuntimeConfig
    lifecycle: LifecycleStatus
    metrics: Metrics
    metadata: Metadata


@dataclass
class KernelListResult:
    """Search result with total count and pagination info for kernels."""

    items: list[KernelInfo]
    total_count: int
    has_next_page: bool
    has_previous_page: bool


# ========== Scheduling History Types ==========


class KernelSchedulingPhase(StrEnum):
    PREPARING = "PREPARING"
    PULLING = "PULLING"
    PREPARED = "PREPARED"
    CREATING = "CREATING"
    RUNNING = "RUNNING"
    TERMINATING = "TERMINATING"
    TERMINATED = "TERMINATED"


@dataclass
class KernelSchedulingHistoryData:
    """Domain model for kernel scheduling history."""

    id: UUID
    kernel_id: KernelId
    session_id: SessionId

    phase: str  # ScheduleType value
    from_status: KernelSchedulingPhase | None
    to_status: KernelSchedulingPhase | None

    result: SchedulingResult
    error_code: str | None
    message: str

    attempts: int
    created_at: datetime
    updated_at: datetime


@dataclass
class KernelSchedulingHistoryListResult:
    """Search result with pagination for kernel scheduling history."""

    items: list[KernelSchedulingHistoryData]
    total_count: int
    has_next_page: bool
    has_previous_page: bool



from collections.abc import Sequence
from dataclasses import dataclass, field
from typing import override

from ai.backend.manager.actions.action import BaseActionResult
from ai.backend.manager.actions.types import ActionOperationType
from ai.backend.manager.data.permission.object_permission import (
    ObjectPermissionCreateInputBeforeRoleCreation,
)
from ai.backend.manager.data.permission.role import RoleData
from ai.backend.manager.models.rbac_models.role import RoleRow
from ai.backend.manager.repositories.base.creator import Creator
from ai.backend.manager.services.permission_contoller.actions.base import RoleAction


@dataclass
class CreateRoleAction(RoleAction):
    creator: Creator[RoleRow]
    object_permissions: Sequence[ObjectPermissionCreateInputBeforeRoleCreation] = field(
        default_factory=tuple
    )

    @override
    def entity_id(self) -> str | None:
        return None

    @override
    @classmethod
    def operation_type(cls) -> ActionOperationType:
        return ActionOperationType.CREATE


@dataclass
class CreateRoleActionResult(BaseActionResult):
    data: RoleData

    @override
    def entity_id(self) -> str | None:
        return str(self.data.id) if self.data else None



"""Session-specific recorder module for scheduler coordinator.

This module provides session-specialized recorder types.
For generic recorder types, import directly from sokovan.recorder.
"""

from .context import SessionRecorderContext
from .recorder import SessionTransitionRecorder

__all__ = [
    "SessionRecorderContext",
    "SessionTransitionRecorder",
]
